#ifndef lint
static char rcsid[] =
	"$Id: barm.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Batch queue removal program
 *
 * barm [-H host] [-Q queue[@host] ] [user...] [job...]
 *
 * -H specify host (all - specify all hosts known to batch)
 * -Q specify queue (all - specify all queues known to batch)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "pathnames.h"
#include "protocol.h"
#include "reply.h"

static char *	use[] = {
	"barm -h [-Q queue] [-H host] <userid|jobid>",
	"\"Barm\" removes a batch job from a queue on a specified host.",
	"",
	"Options:",
	"-h		This help message.",
#ifdef USE_PORT_SELECTION
	"-L [string]	Local Unix socket to connect",
	"-R [int]	Remote port number to connect",
#endif
	"-Q [string]	Specify queue for job removal (or \"all\").",
	"-H [string]   	Specify remote host for job removal (or \"all\").",
	"-V		Print version information and exit",
	"",
	"$Revision: 4.0 $",
	NULL
};

Bool	dflag = FALSE;

int	s;					/* Socket fd */

static struct	passwd pw;

UIDTYPE	myuid;
GIDTYPE	mygid;

static int	request_parse P((char *queue, char *host, int ac, char **av, Bool isuserhost));
static int	queue_remove_job P((char *tqueue, char *thost, int ac, char **av));

extern int	check_group P((char *user, UIDTYPE, GIDTYPE, char *queue));

extern char *	optarg;				/* Ptr to option argument */
extern int	optopt;				/* Current option */ 
extern int	optind;				/* Argument idx */
extern int	opterr;				/* Error handling by getopt */


static SIGRTN	cleanup P((int));


main(argc, argv)
int	argc;
char **	argv;
{

	char *	cp;
	char *	host = NULL;			/* Host of current interest */
	char *	defhost = NULL;			/* Default host to query */
	char *	queue = NULL;			/* Queue of current interest */
	int	option;				/* Option character */

	struct passwd *	pass;			/* Password structure data */


	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
		(void) signal(SIGHUP, cleanup);

	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void) signal(SIGINT, cleanup);

	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
		(void) signal(SIGTERM, cleanup);

	progname = argv[0];			/* Assign the program name */

	myuid = getuid();
	mygid = getgid();

	if (gethostname(lhost, MAXHOSTNAMELEN) < 0) {
		(void) fprintf(stderr, "gethostname fails\n");
		exit(1);
	}

	/*
	 * Check first for a "BATCH" environment variable -
	 * backwards compatable, but clashes with Splus driver envvar.
	 */
	if ((cp = getenv("BATCH")) != NULL)
		queue = estrdup(cp);
	else
		queue = estrdup(_DEF_QUEUE);

	/*
	 * Check second for a "BATCHQUEUE" environment variable.
	 * If there is one, use it.
	 */
	if ((cp = getenv("BATCHQUEUE")) != NULL) {
		free(queue);
		queue = estrdup(cp);
	}

	/*
	 * Always start with the default host.
	 */
	defhost = _DEF_HOST;				

	/*
	 * ... modifiy if the default host is also the localhost.
	 */
	if (islocalhost(defhost) == 0)
		defhost = lhost;
	
	/*
	 * Parse options
	 */
	while ((option = getopt(argc, argv, "dhL:R:H:Q:V")) != EOF)
		switch((char) option) {
#ifdef DEBUG
		case 'd':
			dflag = TRUE;
			break;
#endif
		case 'h':			/* Help */
			usage(use);
			exit(1);
			break;
#ifdef USE_PORT_SELECTION
		case 'L':			/* Local port or socket */
			socketname = optarg;
			break;
		case 'R':			/* Remote port */
			remote_port = atoi(optarg);
			break;
#endif
		case 'Q':			/* Queue name */
			queue = optarg;
			break;
		case 'H':			/* Host name */
			host = optarg;
			break;
		case 'V':
			dump_version(printsf, 0);
			exit(0);
			break;
		case '?':			/* Unknown option */
			if (opterr == 0)
				blog(BLOG_WARNING,
				     "internal: bad flag \"-%c\"", optopt);
			usage(use);
			break;
		default:			/* Unhandled option */
			blog(BLOG_WARNING,
			     "internal: no action for \"-%c\" option",
			     (char) option);
			usage(use);
			break;
		}


	/*
	 * Queue may be queue@host - if so, break it apart
	 */
	if ((cp = strchr(queue, '@')) != NULL) {
		*cp++ = '\0';
		host = cp;
	}

	/* If host is not set, use the default host */
	if (!host)
		host = defhost;
	
	/*
	 * We're running as suid root, and maybe sgid daemon, so we need
	 * to get the *real* uid to handle security on the request.
	 */
	if ((pass = getpwuid(myuid)) == NULL) {
		blog(BLOG_FATAL, epid, myuid);
		cleanup(0);
	} else {
		pw.pw_name = estrdup(pass->pw_name);
		pw.pw_uid = pass->pw_uid;
		pw.pw_gid = pass->pw_gid;
	}

	/*
	 * We are expecting arguments, complain and exit if there are none
	 */
	if (optind == argc) {
		(void) fprintf(stderr,
			       "%s: please specify a user name or a job id",
			       progname);
		usage(use);
		exit(0);
	}


	/* If host is not set, use the default host */
	if (!host)
		host = defhost;

	/*
	 * Parse the given queue and host
	 */
	(void) request_parse(queue, host, argc, argv,
			     (host == defhost) ? FALSE : TRUE);

	exit(0);

	/*NOTREACHED*/
}


/*
 * Check the queue and/or host requests for "all" and other special
 * circumstances (load balancing).  Break into single requests and 
 * pass to the function that calls the host.
 */
static int
request_parse(queue, host, ac, av, isuserhost)
char *	queue;
char *	host;
int 	ac;
char **	av;
Bool	isuserhost;
{

	char *	ho;				/* Next host */
	char *	qu;				/* Next queue */
	char	bbuf[BUFFER];
	char	pbuf[BUFFER];
	int	status;				/* Return status */
#ifdef USE_DELETE_REMOTE
	Bool	loadb;				/* Load balance? */
	char *	bp;
	char *	rm, *rq;			/* Remote host and queue */
#endif

	extern	int optind;
	

	if (strcasecmp(queue, all) == 0) {		/* All queues */
		if (strcasecmp(host, all) == 0) {	/* All hosts */
			while ((ho = getbhent()) != NULL) {
				(void) printf("%s:\n", ho); /* Print host */
				while ((qu = getquent()) != NULL) {
					/* Check group access to this queue */
					if (check_group(pw.pw_name, myuid,
							mygid, qu) < 0) {

						(void) printf(ques, qu,
			      "not a member of the appropriate group(s)\n");
						continue;
					}
					(void) queue_remove_job(qu, ho,	ac, av);
					(void) fputc('\n', stdout);
				}
				endbaent();
				(void) fputc('\n', stdout);
			}
			/* endbhent(); */
		} else {				/* Just all queues */
			while ((qu = getquent()) != NULL) {
				if (check_group(pw.pw_name, myuid,
						mygid, qu) < 0) {

					(void) printf(ques, qu,
			      "not a member of the appropriate group(s)\n");
					continue;
				}
				(void) queue_remove_job(qu, host, ac, av);
				(void) fputc('\n', stdout);
			}
			endbaent();
		}

	} else {					/* Single queue */

		/*
		 * Look for the specified queue name
		 */
		if ((status = bgetent(bbuf, queue)) < 0) {
			blog(BLOG_FATAL, "cannot open queue description file");
			cleanup(0);
		} else if (status == 0) {
			blog(BLOG_WARNING, enotq, queue);
			return (-1);
		}

		/*
		 * Reset the queue name to the canonical (first) name
		 */
		qu = bnamget(bbuf);

		/*
		 * Check for the correct group access before proceeding
		 */
		if (check_group(pw.pw_name, myuid, mygid, qu) < 0) {

			(void) printf(ques, qu,
				      "not a member of the appropriate group(s)\n");
			return (-1);
		}

#ifdef USE_DELETE_REMOTE
		loadb = bgetflag("lb");
		bp = pbuf;
		rm = bgetstr("rm", &bp);
		rq = bgetstr("rq", &bp);
#endif

		/*
		 * User specified a single host
		 */
		if (isuserhost == TRUE && strcasecmp(host, all)) {

			(void) queue_remove_job(qu, host, ac, av);

		/*
		 * If we are to examine all hosts, or we load balance
		 */
		} else if (strcasecmp(host, all) == 0

#ifdef USE_DELETE_REMOTE
			   || loadb == TRUE) {
#else
			                   ) {
#endif /* USE_DELETE_REMOTE */

			while ((ho = getbhent()) != NULL) {
				(void) printf("%s:\n", ho);
				(void) queue_remove_job(qu, ho, ac, av);
				(void) fputc('\n', stdout);
			}
			/* endbhent(); */

		/*
		 * else, if we have a remote queue...
		 */
		} else

#ifdef USE_DELETE_REMOTE

			if (rm != NULL && strcasecmp(rm, lhost) != 0) {

		(void) printf("%s:\n", lhost);
		(void) queue_remove_job(qu, lhost, ac, av);

		(void) printf("\n%s:\n", rm);
		(void) queue_remove_job(rq, rm, ac, av);

		/*
		 * else, just our default host...
		 */
	        } else

#endif /* USE_DELETE_REMOTE */

		       {
			(void) queue_remove_job(qu, host, ac, av);
		}
	}

	return 0;
}



/*
 * Attempt to remove the job from the queue on the (possibly remote) host.
 * through the host_query() function.
 */
static int
queue_remove_job(tqueue, thost, ac, av)
char *	tqueue;
char *	thost;
int	ac;
char **	av;
{

	int 	i;
	char *	cp;
	char	list[BUFFER];

	extern	int optind;


	cp = list;

	/* First copy in the queue name, */
	(void) strcpy(list, tqueue);	

	/* ... then the current user name... */
	cp += strlen(cp);
	*cp++ = ' ';
	(void) strcpy(cp, pw.pw_name);

	/* then all the jobs to be removed. */
	for (i = optind; ac > i; i++) {
		cp += strlen(cp);
		*cp++ = ' ';
		(void) strcpy(cp, av[i]);
#ifdef DEBUG
		if (dflag)
			printf("Command string after arg %d:\"%s\"\n",
			       i, list);
#endif

	}
	cp += strlen(cp);
	*cp++ = '\0';

#ifdef DEBUG
	if (dflag)
		printf("Command string sent:\"%s\"\n", list);
#endif

	/*
	 * try to remove the job(s)
	 */
	return host_query(thost, QC_REM, list);

}


/*
 * Client function to catch the OK return of the host_query call.
 */
int
reply_ok(fp, fmt, queue)
FILE *	fp;
char *	fmt;
char *	queue;
{

	char recvbuf[BUFSIZ];

	while (fgets(recvbuf, sizeof(recvbuf), fp) != NULL) {
		(void) fputs(recvbuf, stdout);
	}

	return 0;
}



/*
 * Client function to catch the MT return of the host_query call.
 */
int
reply_mt(fp, fmt, queue)
FILE *	fp;
char *	fmt;
char *	queue;
{
	return 0;
}


static SIGRTN
cleanup(sig)
int	sig;
{

	(void) close(s);
	exit(1);
}

