#ifndef lint
static char rcsid[] =
	"$Id: bid.c,v 4.0 1993/09/17 00:41:04 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * bid.c - calculate and return current bid.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#include <sys/socket.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include "batch.h"
#include "batchd.h"
#include "bid.h"
#include "flags.h"

#ifdef USE_TCP_BID
#include "protocol.h"

static Bidpacket *	bid_request P((char *, Queue *));
#else
static int		bid_sendto P((char *, Bidpacket *));
#endif

static void		bid_update P((char *, char *,
				      double [], double, time_t));

static Hostinfo *	gethostinfobyhost P((char *));

Hostinfo *		hostinfo_alloc P((void));

static Bidinfo *	getbidinfobyqueue P((qelem *, char *));
static Bidinfo *	bidinfo_alloc P((qelem *));

extern struct in_addr bcastaddr;		/* Broadcast address */

#define MAXBIDDIFF	(0.5)

extern double	CPUfactor;			/* offset for CPU balance */

/* 
 * For each host, get a bid for the queue of interest.
 * Determine the best bid and return the corresponding hostname.
 */
char *
bid_lowest(qp)
Queue *	qp;
{

	Hostinfo *	hip;		/* Pointer to host list */
	Bidinfo	 *	bip;		/* Bids in an individual host list */
	char *	besthost;
	double	bestbid;
	time_t	current;

	current = time((time_t *) 0);

#ifdef USE_TCP_BID
	char *	bhhost;        /* Current host */

	Bidpacket *	bpp;

	/*
	 * Send out requests for information on the different machines
	 * to update our internal load balance database.  If no bids are
	 * returned, or errors occur, specify the local machine for jobs.
	 */
	while ((bhhost = getbhent()) != NULL) {
		
		if ((bpp = bid_request(bhhost, qp)) != NULL) {
			bid_update(bhhost, bpp->bp_name,
				   bpp->bp_loadave, bpp->bp_bid,
				   bpp->bp_stamp);
		} else
			/* bad bid_request() */ ;
	}

	/* endbhent(); */
#endif

	/*
	 * If nothing better comes along, use the current host.
	 */
	bestbid = bid_calculate(qp);
	besthost = lhost;

	/*
	 * For each host, check for information on the queue requested.
	 * Then compare that bid against the best bid.  Keep a pointer to
	 * best host so far.
	 */
	ITER(hip, Hostinfo, &hosts) {
		if ((bip = getbidinfobyqueue(&(hip->hi_bids), qp->q_name)) != NULL) {

			if (current - bip->bi_stamp > btimer) {
				dbg((BLOG_LOAD,
				     "bid for %s@%s (%.2lf) expired %s ago",
				     qp->q_name, hip->hi_name, bip->bi_bid,
				     prttime(current - bip->bi_stamp - btimer)));
				continue;
			} else {
				dbg((BLOG_LOAD,
				     "bid for %s@%s is %.2lf",
				     qp->q_name, hip->hi_name, bip->bi_bid));
			}

			if (bip->bi_bid < bestbid) {
				bestbid = bip->bi_bid;
				besthost = hip->hi_name;
			}
		}
	}
	
	dbg((BLOG_LOAD, "bestbid is for %s", besthost));

	return besthost;
	
}


#ifdef USE_TCP_BID

/*
 * Get the bid for the named queue from the named host.  We check to
 * see if the host is remote or not.  If the named host is the
 * localhost, then feed the information by hand.  We can't call
 * ourselves; we'd block forever!.  Use the rq variable for the queue
 * name if it is set.  Update the bids within this call.
 */
static Bidpacket *
bid_request(thost, qp)
char *	thost;
Queue *	qp;
{

	char  *	queue, *cp;
	int	i, s;
	int	retries = 3;
	unsigned int n;
	static Bidpacket packetbuf;
	int	rfd;


	/*
	 * Update ourselves...
	 */
	if (islocalhost(thost) == 0) {
		(void) strncpy(packetbuf.bp_name, qp->q_name, MAXQUEUENAMELEN);
		(void) memcpy((voidp) packetbuf.bp_loadave,
			      (voidp) loadave,
			      sizeof(loadave));
		packetbuf.bp_bid = bid_calculate(qp);
		packetbuf.bp_stamp = time((time_t *) 0);

		return ((Bidpacket *) &packetbuf);
	}


	/*
	 * If rq is set, use it for the queue name when calling the host
	 */
	if (qp->q_rname != NULL)
		queue = qp->q_rname;
	else
		queue = qp->q_name;

	if ((s = getsocket(thost, service, tcp,	
			   (u_long *) 0, (u_short *) 0)) < 0)
		return ((Bidpacket *) NULL);

	/*
	 * In order to do the communications correctly we really ought to be
	 * forking this off, but we can't and get the information back to
	 * the main daemon.  Therefore  writes and reads must be non blocking
	 * so that if two batchds are in the same section of code, they won't
	 * get into a "deadly embrace" waiting for each other.
	 *
	 * Another way to do this would be to set up a select for the
	 * socket with a time out...
	 */
	if (fcntl(s, F_SETFL, FNDELAY) < 0)
		blog(BLOG_ERR, "bid_request() socket");

	dbg(((BLOG_LOAD | BLOG_XFER), "requesting bid for %s@%s (remote: %s)",
	    qp->q_name, thost, queue));

	(void) sprintf(line2, "%c%s\n", QC_BID, queue);
	
	n = strlen(line2);

	for (i = 1; i <= retries; i++) {
		if (write(s, line2, n) != n) {
			if (errno == EWOULDBLOCK) {
				/*
				 * If we get a would block, wait for a few,
				 * then try to write (read) again. After
				 * a certain number of retries give up.
				 */
				dbg(((BLOG_LOAD | BLOG_XFER),
				     "write would block (%d)", i));
				(void) sleep((unsigned)i);
				continue;
				
			} else {
				blog(BLOG_ERR, elost, "bid_request(): write");
				(void) close(s);
				return ((Bidpacket *) NULL);
			}
		} else {
			goto host_read;
		}
	}

	blog(BLOG_WARNING, "gave up on socket write to %s", thost);
	(void) close(s);
	return((Bidpacket *) NULL);

 host_read:

	/*
	 * Sleep at least one second
	 */
	(void) sleep(1);

	for (i = 1; i <= retries; i++) {
		if (read(s, line2, 1) != 1) {
			if (errno == EWOULDBLOCK) {
				dbg(((BLOG_LOAD | BLOG_XFER),
				     "read would block (%d)", i));
				(void) sleep((unsigned)i);
				continue;
				
			} else {
				blog(BLOG_ERR, elost, "bid_request(): read");
				(void) close(s);
				return ((Bidpacket *) NULL);
			}
		} else {
			goto host_switch;
		}
	}	

	/*
	 * If we get a would block, that means the batchd on
	 * the remote host is doing something.  Just give up.
	 */
	blog(BLOG_WARNING, "gave up on socket read from %s", thost);
	(void) close(s);
	return((Bidpacket *) NULL);

 host_switch:

	dbg(((BLOG_LOAD | BLOG_XFER), "receiving bid for %s@%s",
	    qp->q_name, thost));

	switch ((int) line2[0]) {
	case QR_OK:

		/*
		 * Scan in the remote host name the three load numbers and
		 * the bid.  All doubles.
		 */
		if (read(s, line2, sizeof(line2) > 0) {
				
			line2[strlen(line2) - 1] = '\0';

			(void) sscanf(line2, BID_FMT,
				      &(packetbuf.bp_loadave[0]),
				      &(packetbuf.bp_loadave[1]),
				      &(packetbuf.bp_loadave[2]),
				      &(packetbuf.bp_bid),
				      &(packetbuf.bp_stamp));
			
			/*
			 * Although we asked for the information under the
			 * remote name for the queue, record the information
			 * under our name for the queue.
			 */
			(void) strncpy(packetbuf.bp_name, qp->q_name, MAXQUEUENAMELEN);
			packetbuf.bp_stamp = time((time_t *) 0);
		}  
		(void) close (s);
		return( (Bidpacket *) &packetbuf);
		break;
			       
	case QR_ENOIMP:
		dbg((BLOG_LOAD, "cannot get info from %s", thost));
		break;

	case QR_MT:
		dbg((BLOG_LOAD, "%s@%s (remote: %s) doesn't load balance",
		    qp->q_name, thost, queue));
		break;

	case QR_ENOENT:
		dbg((BLOG_LOAD, "%s@%s (remote: %s) doesn't exist",
		    qp->q_name, thost, queue));
		break;

	case QR_EPERM:
		break;

	case QR_EHPERM:
		dbg((BLOG_LOAD, "%s doesn't allow connctions by us", thost));
		break;
	default:
		blog(BLOG_WARNING,
		      "bid_request(): bogus answer from %s", thost);
		break;
	}

	(void) close (s);
	return ((Bidpacket *) NULL);

}


#else /* USE_TCP_BID */


/*
 * Send a bid for the specified queue to a particular host, or all hosts
 * if thost == NULL.
 * This is called on from in queue_scan() if there are jobs in the queue,
 * or the state of the queue has changed, or the queue has reloaded 
 * ("forced").
 */ 
int
bid_send(thost, qp, current, force)
char *	thost;
Queue *	qp;
time_t	current;
int	force;
{

	char *	bhhost;
	Bool	make_bid = FALSE;
	Hostinfo *	lhp;
	Bidinfo	 *	lbp;
	Bidpacket	packetbuf;
	double		bid;


	bid = bid_calculate(qp);

	/*
	 * If the bid for this queue is substantially different,
	 * or its been longer than btimer, send out the bid.
	 * else its just not worth the effort.
	 */
	if ((lhp = gethostinfobyhost(lhost)) != NULL) {
		if ((lbp = getbidinfobyqueue(&(lhp->hi_bids), qp->q_name)) != NULL) {
			/*
			 * If the timestamp is greater than half the ttl
			 * for a bid, make a new bid.
			 */
			if (current - lbp->bi_stamp > (btimer / 2)) {
				dbg((BLOG_LOAD,
				     "%s: bid half expired %ld > %ld",
				     qp->q_name,
				     current - lbp->bi_stamp, (btimer/2)));
				make_bid = TRUE;
			}
			if (ABS(bid - lbp->bi_bid) > MAXBIDDIFF) {
				dbg((BLOG_LOAD,
				     "%s: bid outdated |%.2f| > %.2f",
				     qp->q_name,
				     bid - lbp->bi_bid, MAXBIDDIFF));
				make_bid = TRUE;
			}
		} else {
			dbg((BLOG_LOAD, "%s: no bid for queue", qp->q_name));

			make_bid = TRUE;	/* No queue entry yet */
		}
	} else {
		dbg((BLOG_LOAD, "%s: no bid for host: %s", qp->q_name, lhost));
		make_bid = TRUE;		/* No host entry yet */
	}

	if (!make_bid && !force) {
		dbg((BLOG_LOAD, "%s: bid remains %.2lf",
		    qp->q_name, lbp->bi_bid));
		return 1;
	} else {
		dbg((BLOG_LOAD, "%s: constructing bid of %.2lf",
		     qp->q_name, bid));
	}


	(void) strncpy(packetbuf.bp_name, qp->q_name, MAXQUEUENAMELEN);

	packetbuf.bp_boot	= boottime;
	packetbuf.bp_stamp	= current;

	/*
	 * If the daemon has been told to reset then advertize the fact.
	 * It's a signal to the other hosts that we'd like to hear their
	 * bids on this queue.
	 * This happens when the daemon starts up or bactchcap is reread.
	 */
	if (queue_get_flag(qp, RESET) == 0) {
		packetbuf.bp_start = (time_t) (-1);
	} else {
		packetbuf.bp_start = starttime;
	}

	packetbuf.bp_bid = bid;
	(void) memcpy((voidp) packetbuf.bp_loadave,
		      (voidp) loadave,
		      sizeof(loadave));

	/*
	 * Send out information to the different machines
	 * to update their internal load balance database.
	 */
	if (thost == NULL) {
		/*
		 * If multicasting, fake a recv on the local host and
		 * multicast on the wire to send to the rest of the hosts.
		 */
   		if (Bflag || Mflag) {
			(void) bid_receive(lhost, &packetbuf, current);
			(void) bid_sendto(thost, &packetbuf);
		} else {
			while ((bhhost = getbhent()) != NULL) {
				if (islocalhost(bhhost) == 0)
					(void) bid_receive(lhost, &packetbuf, current);
				else
					(void) bid_sendto(bhhost, &packetbuf);
			}
			/* endbhent(); */
		}

	/* Don't send a packet to ourself */
	} else if (islocalhost(thost) != 0) {

		/*
		 * Don't bother to send a packet to this host if we don't
		 * have it in the hosts.batchd.
		 */
		while ((bhhost = getbhent()) != NULL) {
			if (strcasecmp(bhhost, thost) == 0) {
				(void) bid_sendto(thost, &packetbuf);
				endbhent();
				break;
			}
		}
		/* endbhent(); */
	}

	return 0;
}


static int
bid_sendto(thost, bpp)
char *	thost;
Bidpacket *	bpp;
{

	Hostinfo *	hip;
	struct hostent *	hp;
	struct sockaddr_in	to;


	/* Zero out the structure */
	(void) memset((voidp) &to, 0, sizeof (to));

	/* Change from host byte order (back) to network byte order */
	to.sin_port = htons(udp_port);

	/* Internet socket */
	to.sin_family = AF_INET;

	/* If [multi|broad]cast and sending to all, */
	if ((Bflag || Mflag) && !thost) {
		/* Multicast address to send */
		to.sin_addr.s_addr = bcastaddr.s_addr;
		thost = "broadcast";
	} else {
		
		/* else sending to one host only, */
		if ((hip = gethostinfobyhost(thost)) == NULL) {
			hip = hostinfo_alloc();
			hip->hi_name = estrdup(thost);
		}

		if (hip->hi_addr == 0) {

			h_errno = 0;
			if ((hp = gethostbyname(thost)) == NULL) {
				if (h_errno == HOST_NOT_FOUND ||
				    h_errno == NO_DATA) {
					blog(BLOG_WARNING, eaddr, thost);
					return (-1);
				}
				return (0);		/* try again later */
			} else {
				(void) memcpy((voidp) &to.sin_addr,
					      (voidp) hp->h_addr,
					      hp->h_length);

				/* Save the address to the cache */
				if (use_cache)
					hip->hi_addr = to.sin_addr.s_addr;
			}

		} else {
			to.sin_addr.s_addr = hip->hi_addr;
		}
	}

	dbg((BLOG_LOAD, "sending bid to %s (%s)",
	    thost, inet_ntoa(to.sin_addr)));

	/* ...LandShark... */
	if (sendto(fudp, (char *) bpp, sizeof(Bidpacket), 0,
		   (struct sockaddr *) &to, sizeof(to)) < 0)
		blog(BLOG_ERR, "sendto %s failed", inet_ntoa(to.sin_addr));
#ifdef DEBUG
	else
		dbg((BLOG_LOAD, "sendto %s succeeds", inet_ntoa(to.sin_addr)));
#endif

	return 0;
}


/*
 * "Receive" and store bid information from a bid structure;
 */
int
bid_receive(rhost, bpp, current)
char *	rhost;
Bidpacket *	bpp;
time_t	current;
{

	Queue *	qp;
	char *	queue = NULL;


	/*
	 * If we can't locate the queue that came on on the udp packet
	 * it probably was a bum message.
	 */
	ITER(qp, Queue, &queues) {
		
		/*
		 * If this is an empty queue then skip.
		 */
		if (qp->q_name == NULL)
			continue;

		/*
		 * If there is a different remote queue name for this
		 * queue entry, check it against the recv'd queue name
		 */
		if (qp->q_rname != NULL) {
			if (strcasecmp(qp->q_rname, bpp->bp_name) == 0) {
				/*
				 * If that was it, store the bid information
				 * under our name for the queue.
				 */
				queue = qp->q_name;
				break;
			}
			
		} else {
			/* ... else check the normal name */
			if (strcasecmp(qp->q_name, bpp->bp_name) == 0) {
				queue = qp->q_name;
				break;
			}
		}
	}


	/*
	 * If no queue found, problem packet.
	 */
	if (queue == NULL) {
		blog(BLOG_WARNING, "bogus bid packet from \"%s@%s\"", 
		    bpp->bp_name, rhost);
		return (-1);

	} else {

		dbg((BLOG_LOAD, "bid from %s@%s (queue: %s)",
		    bpp->bp_name, rhost, queue));

	}

	bid_update(rhost, queue, bpp->bp_loadave, bpp->bp_bid, bpp->bp_stamp);

	/*
	 * A starttime of -1 to indicates that the remote host just came up.
	 * Send a bid for this queue back to the originating host if we are
	 * contributing to load balancing this queue.
	 */
	if (qp->q_loadb && bpp->bp_start == -1) {
		dbg((BLOG_LOAD, "sending bid to reset host %s@%s (queue: %s)",
		    bpp->bp_name, rhost, queue));

		(void) bid_send(rhost, qp, current, 0);
	}

	return 0;
	
}


#endif /* USE_TCP_BID */


/* 
 * Update the information structure for the given hostname and queue.
 * The name that the host is referenced by is the remote hosts idea of
 * it's own name; the name that comes back from the gethostname() call.
 * The refereced queue name is our idea of the queue name.  The rq batchcap
 * variable (in qp->q_rname) is used if appropriate.
 */
static void
bid_update(rhost, queue, lav, bid, stamp)
char *	rhost;
char *	queue;
double	lav[3];
double	bid;
time_t	stamp;
{
	
	Hostinfo *	hip;
	Bidinfo	 *	bip;
	
	
	if ((hip = gethostinfobyhost(rhost)) == NULL) {
		hip = hostinfo_alloc();
		hip->hi_name = estrdup(rhost);
	}

	(void) memcpy((voidp) hip->hi_loadave,
		      (voidp) lav,
		      sizeof(loadave));

	hip->hi_stamp = stamp;
	
	if ((bip = getbidinfobyqueue(&(hip->hi_bids), queue)) == NULL) {
		bip = bidinfo_alloc(&(hip->hi_bids));
		bip->bi_name = estrdup(queue);
	}

	bip->bi_bid = bid;
	bip->bi_stamp = stamp;
	bip->bi_count++;
	
	dbg((BLOG_LOAD, "updated host %s@%s with bid %.2lf",
	    queue, hip->hi_name, bid));
	
	return;
}


/*
 * Get a bidding host by name.
 */
static Hostinfo *
gethostinfobyhost(hostname)
char *	hostname;
{
	
	Hostinfo *	hip;
	
	/*
	 * Scan the hosts structure till we find the rhost name.  If we
	 * don't find the name return null.
	 */
	ITER(hip, Hostinfo, &hosts) {
		/*
		 * If this is an empty entry then skip it
		 */
		if (hip->hi_name == NULL)
			continue;
		
		if (strcasecmp(hostname, hip->hi_name) == 0) {
			return (Hostinfo *) hip;
		}
	}
	return (Hostinfo *) NULL;
}

Hostinfo *
hostinfo_alloc()
{

	Hostinfo *	hnew;


	/*
	 * If we fall off the end, make space for a new Hostinfo and
	 * copy in the name.
	 */
	dbg((BLOG_ALLOC, "allocating Hostinfo structure"));

	hnew = ALLOC(Hostinfo);
	
	qinfo.i_hostlist++;

	hnew->hi_bids.q_forw = &(hnew->hi_bids);
	hnew->hi_bids.q_back = &(hnew->hi_bids);

	ins_tail_que(&hnew->hi_hosts, &hosts);

	return hnew;

}

/*
 * Given a pointer to a Hostinfo structure, get a queue bid by name.
 */
static Bidinfo *
getbidinfobyqueue(bids, queue)
qelem *	bids;
char *	queue;
{
	
	Bidinfo *	bip;
	
	ITER(bip, Bidinfo, bids) {
		
		/*
		 * If this is an empty entry then skip it
		 */
		if (bip->bi_name == NULL)
			continue;
		
		if (strcasecmp(queue, bip->bi_name) == 0) {
			return (Bidinfo *) bip;
		}
	}

	return (Bidinfo *) NULL;
}

static Bidinfo *
bidinfo_alloc(bids)
qelem *	bids;
{

	Bidinfo *	bnew;


	/*
	 * Add a new bid to the list.
	 */
	dbg((BLOG_ALLOC, "allocating Bidinfo structure"));

	bnew = ALLOC(Bidinfo);
	
	qinfo.i_bidlist++;

	ins_tail_que(&bnew->bi_bids, bids);

	return bnew;
	
}


/* 
 * Do the acutal bid calculation in floating point math.
 *
 * If q_count is used instead of q_nexec, there is a flip flop of 
 * jobs between hosts, because the current host then always has a
 * higher load than any of the other hosts, so it sends it off again.
 * queued jobs don't count, cus they'll be passed around anyway.
 * 
 */
double
bid_calculate(qp)
Queue *	qp;
{

	int	run, que;
	double	bid	= 0.0;


	if (qp == NULL) {
		blog(BLOG_ERR, "bid_calculate called with a NULL qp");
		return (double) (-1);
	}

	/*
	 * The multiplicative factors are arbitrary, but indended to
	 * indicate the severity of impact on the machine queue,
	 * and the likelihood that the job will start.
	 */
	if (queue_get_flag(qp, STOPPED) == 0)
		run = 3;
	else if (queue_get_flag(qp,(SUSPEND | INTERRUPT | TIMEOUT)) == 0)
		run = 2;
	else if (queue_get_flag(qp, RUNNING) == 0)
		run = 1;

	if (queue_get_flag(qp, DISABLED) == 0)
		que = 3;
	else if (queue_get_flag(qp, WAITING) == 0)
		que = 2;
	else if (queue_get_flag(qp, QUEUING) == 0)
		que = 1;

	/* 
	 * Multipy the factors together - the lower the bid the better.
	 */
	bid = 1.0;
	bid *= (loadave[LOAD_5MIN] + 1.0);
	bid *= (qp->q_nexec + 1.0);

	/*
	 * If the queue does not have an empty slot increase the bid.
	 */
	bid *= ( qp->q_nexec == qp->q_maxexec ) ? 2.0 : 1.0;

	bid *= run * que;

	/*
	 * An attempt at normalization:
	 * Divide through by the relative processor power (default == 1)
	 * I give the CPUfactor in MIPS (+- memory for that machine).
	 */
	bid /= CPUfactor;

	return bid;

}
