#ifndef lint
static char rcsid[] =
	"$Id: event.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */

/*
 * Event handler stuff
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <ctype.h>

#include <limits.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <time.h>

#include "config.h"

#if (defined(I_STDARG) && defined(CAN_VAPROTO)) || defined(__STDC__)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "event.h"

/* Number of free event structures to keep */
#define FREE_EVENTS	5

qelem events = { &events,  &events };		/* Event list */
qelem fevents= { &fevents, &fevents };		/* Free event list */
static int   free_ep = 0;			/* # avail event structs */

static Event *	event_alloc P((void));
static void	event_free P((Event *));


/*
 * Event manager routines
 */


/*
 * Return the time of the next event, or 0 if no pending events.
 * Have a side effect of executing an event if it's time has come.
 */
time_t
event_next()
{

	Event *	ep;
	time_t	current = time((time_t *) 0);

	int	bestid = 0;		/* Next event to be called */
	int	id;			/* Place holder for id number */
	char *	desc;			/* Place holder for description */
	time_t	best = LONG_MAX;	/* Best time so far */
	void	(*funcp)();		/* Place holder ptr to function */
	voidp	argsp;			/* Place holder to argument */

	/*
	 * All of this is dependent on adding things that will be
	 * executed in the *future* - if we try to add something in
	 * the past, it will be missed until the next timeout, whenever
	 * that is.  Time is "stationary" within this function.
	 */

	/*
	 * For all events. (First event *always* exists).
	 */
	ITER(ep, Event, &events) {

		/*
		 * If it is time to do this event, save the function and
		 * it's arguments, splice this ep out of the list, and
		 * free the current event.
		 *
		 * Then evaluate the function + args.
		 */
		if (ep->e_time <= current) {

			funcp = ep->e_funcp;
			argsp = ep->e_argsp;
			desc = ep->e_desc;
			id = ep->e_id;

			dbg((BLOG_EVENT, "event called: #%d \"%s\"",
			    id, desc != NULL ? desc : "(none)"));

			/*
			 * Free up the event structure before we call
			 * the function, the function might allocate a
			 * new structure - we can give this one back.
			 */
			event_free(ep);

			(*funcp)(argsp);

			dbg((BLOG_EVENT, "event returned: #%d", id));

			/*
			 * returning zero says that there might be
			 * more callouts to execute...
			 */
			return 0;

		} else if (ep->e_time > current && ep->e_time < best) {

			best = ep->e_time;
			bestid = ep->e_id;

		}
	}

	dbg((BLOG_EVENT, "next event: #%d in %lds",
	    bestid, bestid ? (best - current) : 0));

	if (best == LONG_MAX) {
		blog(BLOG_WARNING, "no pending events?");
		return (-1);
	} else {
		return (best - current);
	}
	

}


/*
 * add an event to the list, to fire at time "in", with a pointer to
 * a function and it's arguments.
 */
int
event_add(in, funcp, argsp, desc)
time_t	in;
void	(*funcp)V((voidp));
voidp	argsp;
char *	desc;
{

	Bool	found = FALSE;
	char	temp[LINE];
	int	current = time((time_t *)0);
	static int	e_counter = 0;
	Event *	ep, *	enew;
		
	/*
	 * Create/find a new event slot, set it up and splice it in
	 * based on the time it is to be called;
	 */

	enew = event_alloc();
	e_counter++;

	/*
	 * For all events, check that the id number is not in use.
	 */
	do {

		found = FALSE;
		ITER(ep, Event, &events) {
			if (ep->e_id == e_counter) {
				found = TRUE;
				e_counter++;
				break;
			}
		}

	} while (found == TRUE);
			

	enew->e_funcp = funcp;
	enew->e_argsp = argsp;
	enew->e_time  = in + current;
	enew->e_id    = e_counter;

	if (desc != NULL)
		enew->e_desc  = estrdup(desc);
	else
		enew->e_desc = NULL;

	/*
	 * Iterate until the event ahead of us is greater in time
	 * than we are.
	 */
	ITER(ep, Event, &events) {
		if (ep->e_time >= enew->e_time)
			break;
	}

#ifdef DEBUG
	if (ep == HEAD(Event, &events)) {
		dbg((BLOG_EVENT, "schedule event: #%d last", enew->e_id));
	} else {
		dbg((BLOG_EVENT,
		    "schedule event: #%d ahead of %d (%lds < %lds)",
		    enew->e_id, ep->e_id,
		    (enew->e_time - current), (ep->e_time - current)));
	}
#endif

	ins_tail_que(&enew->e_events, &ep->e_events);

	(void) strcpy(temp, ctime(&enew->e_time));
	temp[16] = '\0';

	dbg((BLOG_EVENT, "schedule event: #%d  (\"%s\" at %s)",
	    enew->e_id, enew->e_desc, temp));

	return (enew->e_id);

}


/*
 * Remove a particular event based on it's event id (e_id)
 * and "free" the associated structure.
 */
int
event_remove(id)
int	id;
{

	Event *	ep;

	ITER(ep, Event, &events) {
		if (ep->e_id == id) {
			dbg((BLOG_EVENT, "remove event: %d (%s)",
			    ep->e_id, ep->e_desc));
			event_free(ep);
			return 0;
		}
	}
	return (-1);
}

void
event_dump(func, lvl)
void	(*func)V((int, char *, ...));
int	lvl;
{

	Event *	ep;
	char	temp[LINE];
	time_t	current = time((time_t *) 0);
	

	ITER(ep, Event, &events) {

		(void) strcpy(temp, ctime(&ep->e_time));
		temp[16] = '\0';
		
		(*func)(lvl, "\t#%-4d %-30.30s in %8lds %s",
			ep->e_id, ep->e_desc,
			(ep->e_time - current), temp);
	}

	return;
}


/*
 * Free() or cubby hole an event structure that we have let go.
 */
static void
event_free(ep)
Event *	ep;
{

	/*
	 * Free up the description, if there is one.
	 */
	efree((voidp) ep->e_desc);
	ep->e_desc = NULL;

	rem_que(&ep->e_events);
	/*
	 * Then either save the structure on a free list, or
	 * just free it alltogether.
	 */
	if (free_ep > FREE_EVENTS) {

		free((voidp) ep);

	} else {

		ins_tail_que(&ep->e_events, &fevents);
		free_ep++;
		qinfo.i_feventlist++;
	}
	
}


/*
 * "Gimme a Event structure!"
 */
static Event *
event_alloc()
{

	Event  *	enew;

	if (free_ep > 0) {

		dbg((BLOG_ALLOC, "reusing Event structure"));
		enew = FIRST(Event, &fevents);
		rem_que(&enew->e_events);
		free_ep--;
		qinfo.i_feventlist--;

	} else {

		/*
		 * Make space for a event job entry
		 */
		dbg((BLOG_ALLOC, "allocating Event structure"));

		enew = ALLOC(Event);

		qinfo.i_eventlist++;
	}
	
	return enew;
}


/*
 * Return minutes to event or -1 if event is in the past 
 */
time_t
event_time(eday, ehour)
int	eday;			/* days from now */
int	ehour;			/* absolute hours */
{
	

	int		chour, cday, cmin;
	struct tm *	local_time = localtime(&now);


	cday = local_time->tm_wday;
	chour = local_time->tm_hour;
	cmin = local_time->tm_min;


	if (eday == 0 && ((MINPERHOUR * chour) + cmin) > ehour)
		return (-1);

	/*
	 * OK - we've got an event now that is ahead of us
	 * in time, so math (done in mimutes) is:
	 * days_offset + time_for_event - current_time
	 */
	return (((eday * MINPERDAY) + ehour -
		 ((chour * MINPERHOUR) + cmin)) * SECPERMIN);
}
