#ifndef lint
static char rcsid[] =
	"$Id: queuejob.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */


/*
 * queuejob.c -  re-analyse and modify the state of submission to the queues.
 *	Scan the spooling directories for new jobs.  If appropriate, send
 *	the jobs to remote hosts based on RM entries, or load balancing.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>

#if defined(MALLOC_TRACE) || defined(MALLOC_LEAK)
#include "dmalloc.h"
#else
#include <malloc.h>
#endif

#include <pwd.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>

#include <sys/dir.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "event.h"
#include "flags.h"
#include "jobsort.h"

#define	FREE_JOBS	10

static qelem	fjobs = { &fjobs, &fjobs };	/* Free job list */
static int	free_jp = 0;		/* Number of job structures avail */

static Job *	jarray[MAXJOBS];

static int	queue_clean P((Queue *));
static int	queue_add_jobs P((Queue *, int));
static int	queue_attach_job P((Queue *, Job *));

static int	job_read_control P((Job *, char *));
static int	job_delay_set P((Queue *, Job *));
static Job *	job_alloc P((qelem *));
void		job_free P((Job *));
void		job_sort P((qelem *));

extern void 	do_periodic P((Queue *));
extern void 	do_scan P((Queue *));
extern void	do_delay P((Job *));

extern void 	send_results P((Queue *, Job *, wait_t *, Notify, char *));

#ifdef HAVE_RLIMIT
extern int 	atorlimit P((rlimit_t [], char *));
#endif

extern char *	bid_lowest P((Queue *));


/*
 * Scan the directory for a queue and find all jobs ready to run.  Enter
 * them in the queue, and possibly send them to another host running batchd.
 *
 * Return the following:
 * STATUS_ERROR - 	an error occured in the normal queueing routine.
 * STATUS_DISABLED -	the queue is or became disabled. 
 * STATUS_NOCHANGE -	nothing happened.
 * STATUS_ENABLED -	the queue became enabled or a job was queued.
 * STATUS_PERSIST -	there was a message put into the queue status file.
 */
int
queue_find_jobs(qp, force)
Queue *	qp;
int	force;
{

	Job *	jp;
	Usr *	up;
	int	status = STATUS_NOCHANGE;
	int	rtn = STATUS_NOCHANGE;
	Bool	restat = FALSE;
	Bool	change = FALSE;
	int	ijobs = 0, rjobs = 0;
	time_t	qmtime;



	/*
	 * If this is the first time the queue is being build, clear
	 * out the dead jobs (those that have a cf and an of file).
	 */
	if (queue_get_flag(qp, RESET) == 0) {
		if ((ijobs = queue_add_jobs(qp, FALSE)) < 0)
			return(STATUS_ERROR);
		if ((rjobs = queue_clean(qp)) < 0)
			return(STATUS_ERROR);
		dbg(((BLOG_JOBS | BLOG_QUEUES),
		    "%s: startup: %d found, %d removed",
		    qp->q_name, ijobs, rjobs));
	}

	/* Change is TRUE if queue is to be disabled in some way */
	change = (queue_get_lock(qp, DISABLED) == 0);
	if ((rtn = queue_change_state(qp, change, DISABLED, force))
	    != STATUS_ENABLED) {
		if (change) {
			if (queue_get_lock(qp, STOPPED) == 0 ||
			    queue_get_lock(qp, DRAINING) == 0)
				return (STATUS_PERSIST);
			else
				return (STATUS_DISABLED);
		}
	} else {
		status = rtn;
	}

	change = (((int) (loadave[LOAD_5MIN] + 0.5) > qp->q_loadsched) &&
		  (qp->q_userrc ? (qp->q_uctl & SUSPEND) : TRUE));

	if ((rtn = queue_change_state(qp, change, WAITING, force))
	    != STATUS_NOCHANGE)
		status = rtn;

	if (queue_get_flag(qp, QUEUING) != 0)
		return (status);

	/*
	 * Now check the spooling directory itself.  If it changed,
	 * then something new might have been queued or TOPQ may have
	 * asked for a re-evaluation.
	 */
	if ((qmtime = getmtime(".")) < 0) {
		return (STATUS_ERROR);
	}

	/*
	 * If there hasn't been a change in the directory, go directly
	 * to the code that resends jobs for remote queues and load balances.
	 */
	if (qmtime != qp->q_mtime || restat == TRUE || force == TRUE) {
		blog(BLOG_TRACE, ques, qp->q_name, "scanning spooling area");
		qp->q_mtime = qmtime;
		if ((ijobs = queue_add_jobs(qp, restat)) < 0)
			return(STATUS_ERROR);
		else if (ijobs > 0)
			status = STATUS_PERSIST;
	} else {
		blog(BLOG_TRACE, ques, qp->q_name, "dormant spooling area");
	}

	/* Clear the user data on the number of queued jobs. */
	ITER(up, Usr, &(qp->q_usrs))
		up->u_queued = 0;

	/* Tally the jobs queued in this queue (but not the running jobs?) */
	ITER(jp, Job, &(qp->q_jobs)) {

		/* Don't count the running jobs */
		if (jp->j_state & JOB_RUNNING)
			continue;

		/*
		 * If the cf file of the job has the owner execute bit
		 * set, then it is a topq'ed job, reevaluate this each
		 * time we scan the queue.
		 */
		if (stat(jp->j_file, &stbuf) < 0) {
			blog(BLOG_WARNING, estat, jp->j_file);
		} else {
			if (stbuf.st_mode & 00100) {
				jp->j_state |= JOB_TOPQ;
				jp->j_mtime = stbuf.st_mtime;
			} else {
				jp->j_state &= ~(JOB_TOPQ);
			}
		}

		if ((up = getusrbyid(&(qp->q_usrs), qp->q_ftype, jp)) == NULL)
			blog(BLOG_WARNING, quej, qp->q_name, jp->j_id,
			    "cannot find usr struct for job");
		else
			up->u_queued++;
	}

	/*
	 * Sort the jobs of this queue
	 */
	job_sort(&(qp->q_jobs));

	/*
	 * For each active, unprocessed job in the queue...
	 */
	ITER(jp, Job, &(qp->q_jobs)) {

		/* If it's a running job we skip it. */
		if (jp->j_state & JOB_RUNNING)
			continue;

		dbg((BLOG_JOBS, quej, qp->q_name, jp->j_id, "queueing"));

		/*
		 * Shall we send this job somewhere?
		 * 1) Job is to be attached.
		 * 2) Queue is on a remote host.
		 * 3) Load balancing queue (and no remote host specified).
		 * 4) User specifed remote host.
		 * 5) default: stay on the localhost.
		 *
		 * So, if the user specifies a remote host (the 'R' record),
		 * this overrides the load balancing option.  A remote host
		 * may be "localhost" or the current host.
		 *
		 */
		if (jp->j_state & JOB_ATTACH) {
			if (jp->j_state & JOB_ATTACHED)
				continue;
			if (queue_attach_job(qp, jp) < 0) {
				job_unlink(jp);	/* Job already running */
			} else {
				status = STATUS_PERSIST;
			}
		} else

		if (qp->q_rhost != NULL) {

			blog((BLOG_TRACE | BLOG_XFER),
			    "job goes to remote queueing host %s", qp->q_rhost);

			if (islocalhost(qp->q_rhost) > 0) {
				jp->j_shost = qp->q_rhost;
				
				if (job_send(jp) != RTN_OK)
					blog(BLOG_WARNING, "could not send job");
			} else {
				/*
				 * This host is the "remote" host for the
				 * queue, the job had already been sent here.
				 */
				continue;
			}

		} else if (qp->q_loadb == TRUE && jp->j_rhost == NULL) {

			if ((jp->j_shost = bid_lowest(qp)) == NULL)
				continue; /* return (STATUS_DISABLED); */

			blog((BLOG_TRACE | BLOG_XFER | BLOG_LOAD),
			    "job goes to load balancing host %s",
			    jp->j_shost);

			if (islocalhost(jp->j_shost) > 0) {

				if (job_send(jp) != RTN_OK)
					blog(BLOG_WARNING, "could not send job");

			} else {

				/* Load balancing host was this host */
				continue;
			}
			 
		} else if (jp->j_rhost != NULL) {

			if (islocalhost(jp->j_rhost) > 0) {

				blog((BLOG_TRACE | BLOG_XFER),
				    "job goes to user specified host %s",
				    jp->j_rhost);

				jp->j_shost = jp->j_rhost;
				if (job_send(jp) != RTN_OK)
					blog(BLOG_WARNING, "could not send job");
					
			} else {

				/*
				 * This host is the "remote" host, the job
				 * had already been sent here.
				 */
				continue;
			}
			            
		} else {
			/*
			 * None of the above: implicit local job
			 */
			continue;
		}

	}

	return status;
}


/*
 * This section of code (while loop) reads the files in
 * the spooling directory, and adds to the queue structures.
 */
static int
queue_add_jobs(qp, restat)
Queue *	qp;
int	restat;
{

	DIR *	qdir;
	Job *	jp, *	djp;
	Usr *	up;
	struct direct * d;
	int	count, i;


	/* Clear the "seen" flag. */
	ITER(jp, Job, &(qp->q_jobs))
		jp->j_seen = FALSE;

	/*
	 * Each time the queue is rebuilt, count the number of jobs in the
	 * queue from scratch.
	 */
	qp->q_count = 0;

	if ((qdir = opendir(".")) == NULL) {
		blog(BLOG_ERR, eopen, qp->q_spool);
		return (-1);
	}

	while ((d = readdir(qdir)) != NULL) {

		/* We're only looking for "cf" files */
		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
			continue;

		/*
		 * Scan over all the queued jobs...
		 */
		ITER(jp, Job, &(qp->q_jobs)) {

			/*
			 * If we've found a previously queued job...
			 */
			if (strcmp(d->d_name, jp->j_file) == 0) {

				/* Mark as seen. */
				jp->j_seen = TRUE;

				/* Increment the # of jobs in the queue */
				qp->q_count++;

				/* Break out of the for loop. */
				break;
			}
		}

		/*
		 * If we ran off the end of the table, we didn't
		 * find it as an old job. So add it to the queue structure. 
		 */
		if (jp == HEAD(Job, &(qp->q_jobs))) {

			/* Add a new job to the list. */
			jp = job_alloc(&(qp->q_jobs));

			/* Be certain to set the pid to zero. */
			jp->j_pid = 0;

			/*
			 * If we don't know any better, a job to be run,
			 * but not yet running.
			 */
			jp->j_state |= JOB_RUN;
			jp->j_state &= ~(JOB_RUNNING);

			/* Set pointer back to queue */ 
			jp->j_queue = qp;

			/* Set the default CPU for this job */
   			jp->j_mustrun = qp->q_mustrun;

			/* Copy the directory entry name to the file name. */
			jp->j_file = estrdup(d->d_name);

			/* Copy the priority values */
			jp->j_priority   = qp->q_priority[NICE_LOW];
			jp->j_ndpriority = qp->q_ndpriority;

#ifdef HAVE_RLIMIT
			/* Copy the resoure limits */
			(void) memcpy((voidp) jp->j_rlimit,
				      (voidp) qp->q_rlimit,
				      sizeof(qp->q_rlimit));
#endif

			/*
			 * Read in the control file information.
			 * If there is a problem, that means some aspect
			 * of the user/group information was incorrect.
			 * That situation would be logged as an error, so
			 * now we remove the job files.
			 */
			if (job_read_control(jp, d->d_name) < 0) {
				job_unlink(jp);
				continue;
			}

			/* Chown the cf file to the owner of the job */
			if (chown(jp->j_file, jp->j_uid, jp->j_gid) < 0)
				blog(BLOG_ERR, echow, jp->j_file);
			
			jp->j_file[0] = 'd';

			/* Chown the df file to the owner of the job */
			if (chown(jp->j_file, jp->j_uid, jp->j_gid) < 0)
				blog(BLOG_ERR, echow, jp->j_file);

			jp->j_file[0] = 'c';

			/*
			 * If we can't get the create date from the job file
			 * set it here using a stat call.
			 */
			if (jp->j_ctime == 0 || jp->j_mtime == 0) {
				if (stat(d->d_name, &stbuf) < 0) {
					blog(BLOG_ERR, estat, d->d_name);
					continue;
				}

				if (jp->j_ctime == 0)
					jp->j_ctime = stbuf.st_ctime;

				if (jp->j_mtime == 0)
					jp->j_mtime = stbuf.st_mtime;

			}
			
			/*
			 * Set an event for the delayed start
			 */
			if (job_delay_set(qp, jp) < 0)
				continue;

			/* Increment the # of jobs in the queue */
			qp->q_count++;

			/* Do our internal accounting */
			if ((up = getusrbyid(&(qp->q_usrs), qp->q_ftype, jp)) == NULL) {
				up = usr_alloc(&(qp->q_usrs));
				if (qp->q_ftype & SORT_UID)
					up->u_uid = jp->j_uid;
				else if (qp->q_ftype & SORT_GID)
					up->u_gid = jp->j_gid;
				else
					blog(BLOG_ERR, "usr type error");
			} 

			/* Set pointer back to usr struct */ 
			jp->j_usr = up;

			/* and tally a submitted job */
			up->u_submit++;

			/*
			 * Tell the user (via the status file) that we
			 * have queued another job.
			 */
			(void) sprintf(line, quej, qp->q_name, jp->j_id,
				       "queued");

			blog(BLOG_STATS, line);
			queue_write_status(qp, line);

		} /* endif */

		/* In both cases mark it as being seen */
		jp->j_seen = TRUE;

	} /* endwhile */

	(void) closedir(qdir);


	/*
	 * Look at our internal table and check that every job we
	 * know about is still in the directory.  If one has been
	 * deleted just remove the job from the table.  This code 
	 * executes if "barm" removes a queued job.
	 *
	 * If the job was running, wait for it to return, and then remove
	 * the job structure from the list in the process_reap() function.
	 *
	 */

	/* Gather pointers to the jobs we want to remove. */
	count = 0;
	ITER(jp, Job, &(qp->q_jobs)) {
		if (jp->j_seen == FALSE && !(jp->j_state & JOB_RUNNING)) {
			jarray[count++] = jp;
		}
	}

	/* Free the dead jobs */
	for (i = 0; i < count; i++)
		job_free(jarray[i]);

	return qp->q_count;
}


/*
 * Clean out the queue.
 * This function must be called *after* the queue is rebuilt in order
 * for it to work correctly.
 */
static int
queue_clean(qp)
Queue *	qp;
{

	int	ijobs, i, count;
	Job *	jp;
	DIR *	qdir;
	struct direct *	d;


	if ((qdir = opendir(".")) == NULL) {
		blog(BLOG_ERR, eopen, qp->q_spool);
		return (-1);
	}

	/* Record the number of jobs initially in the queue. */
	ijobs = qp->q_count;

	/*
	 * For all files in the spooling directory...
	 */
	count = 0;
	while ((d = readdir(qdir)) != NULL) {

		/*
		 * We're only interested in "of" (output files) as those
		 * indicate that a job was running.
		 */
		if (d->d_name[0] != 'o' || d->d_name[1] != 'f')
			continue;
		
		ITER(jp, Job, &(qp->q_jobs)) {

			/* Look for the corresponding "cf" file in the table */
			if (strcmp(d->d_name+1, jp->j_file+1) != 0)
				continue;

			/* Mail the user about the crashed job */
			send_results(qp, jp, (wait_t *) 0,
				     (Notify) (NOTIFY_CRASH | NOTIFY_DEBUG),
				     (qp->q_restart == TRUE &&
				      jp->j_restart == TRUE) ?
				     "restarted" : "*not* restarted");

			if (qp->q_restart == TRUE && jp->j_restart == TRUE) {
				jp->j_state |= JOB_RESTART;
				if (aflag > 0 && !(log_lvl & BLOG_NOFORK))
					account(qp, "boot  %s job restarted:%d",
						qp->q_name, jp->j_id);
				(void) sprintf(line, quej,
					       qp->q_name, jp->j_id,
					       "restarted");

				blog(BLOG_STATS, line);
				queue_write_status(qp, line);
				break;
			} else {

				if (aflag > 0 && !(log_lvl & BLOG_NOFORK))
					account(qp, "boot  %s job removed:%d",
						qp->q_name, jp->j_id);
				(void) sprintf(line, quej,
					       qp->q_name, jp->j_id,
					       "not restarted");

				blog(BLOG_STATS, line);
				queue_write_status(qp, line);
			}

			/* If restart is off, delete the job control files */
			job_unlink(jp);

			/* Gather this job pointer for later removal */
			jarray[count++] = jp;

			qp->q_count--;

			break;
		}


		/*
		 * Clean out output files that don't have control files.
		 * If there was no "cf" file corresponding to the "of" file
		 * we found, remove the "of" and the "df" file.
		 */
		if (jp == HEAD(Job, &(qp->q_jobs))) {

			d->d_name[0] = 'o';		/* necessary ? */

			if  (unlink(d->d_name) < 0) {
				blog(BLOG_ERR, eunlk, d->d_name);
			} else { 
				if (qp->q_supmail & NOTIFY_CRASH)
					send_mail(qp->q_smaddress,
						  "Batchd crash cleanup",
						  "Queue %s: output file \"%s\" deleted",
						  qp->q_name, d->d_name);

				if (qp->q_supwrite & NOTIFY_CRASH)
					send_write(qp->q_swaddress,
						   "Queue %s: output file \"%s\" deleted",
						   qp->q_name, d->d_name);

			}

			/* Remove the data file too */
			d->d_name[0] = 'd';

			if (unlink(d->d_name) < 0 && errno != ENOENT)
				blog(BLOG_ERR, eunlk, d->d_name);

		} /* endif */

	} /* endwhile */

	(void) closedir(qdir);

	/* Free the dead jobs */
	for (i = 0; i < count; i++)
		job_free(jarray[i]);

	if (aflag > 0 && !(log_lvl & BLOG_NOFORK))
		account(qp, "boot  %s queue restarted", qp->q_name);

	/* return the number of jobs removed */
	return (ijobs - qp->q_count);
}



/*
 * A couple of notes about job parameters:
 *
 * All time values are zero if they are unknown.
 *
 * j_ctime:  The time a job was created.  This can be gotten from the 
 *	ctime on the stat() call, or it can be sent in from a time() call
 *	by "batch."  If it is gotten from the file stat(), then the jobs
 *	that are transmitted in are disadvantaged because the files
 *	they reside in are "new" even though the job is quite old.
 *
 * j_mtime:  The time a job file is modified.  This is initially set to
 *	the creation time of the job (see above), but is modified if topq
 *	is used to place a job at the top of a queue.  N.B.: Job files
 *	*are* modified when 'P', 'T' and 'K'  records are written to them,
 *	but that isn't recorded in the j_mtime entry.
 *
 * j_stime: The time the job is started. Set runjob.
 *
 * j_dtime: The delayed start time of a job. Set here only.
 *
 * j_jname: The first argument of the job or stdin.  This may also be set
 * from the kernel proc strucutre when CPU time is evaluated.
 * If the job has no value in the 'J' slot, j_nname is set to NULL.  
 */

/*
 * Read the control file for a job
 */
static int
job_read_control(jp, filename)
Job *	jp;
char *	filename;
{
	
	char	class;
	char	*cp, *djp, *dqp, *dhp;
#ifndef HAVE_FLAT_NAMES
	struct passwd *	pw;
	struct group *	gr;
#endif
	FILE *	cfp = NULL;
	int	id;
	
	if ((cfp = fopen(filename, "r+")) == NULL) {
		blog(BLOG_ERR, eopen, filename);
		return (-1);
	}

#ifdef USE_LOCALBUF	
	(void) setbuf(cfp, iobuf);
#endif

	dbg((BLOG_JOBS, "cf file: \"%s\"", filename));

	while (fgets(line, sizeof(line), cfp) != NULL) {
		
		/* Nuke the newline */
		line[strlen(line) - 1] = '\0';

		class = line[0];
		cp = &line[1];

		dbg((BLOG_JOBS, "got class: %c line: \"%s\"", class, cp));

		switch(class) {
		case 'a':	/* Start *a*t time */
			jp->j_delay.d_string = estrdup(cp);
			if (strchr(cp, '/') != NULL) {
				jp->j_delay.delay_u.d_file = estrdup(cp);
				jp->j_delay.d_type = DELAY_FILE;
				jp->j_state |= JOB_DELAY;
			} else if ((dqp = strchr(cp, '%')) != NULL) {
				*dqp++ = 0;
				if ((dhp = strchr(dqp, '@')) != NULL)
					*dhp++ = 0;
				else
					dhp = NULL;
				
				jp->j_delay.delay_u.d_djob.d_id = atoi(cp);

				if (dqp != NULL)
					jp->j_delay.delay_u.d_djob.d_queue = estrdup(dqp);
				if (dhp != NULL)
					jp->j_delay.delay_u.d_djob.d_host = estrdup(dhp);
				jp->j_delay.d_type = DELAY_JOB;
				jp->j_state |= JOB_DELAY;
			} else 	if ((jp->j_delay.delay_u.d_time = atol(cp)) > 0) {
				jp->j_delay.d_type = DELAY_TIME;
				jp->j_state |= JOB_DELAY;
			} else {
				blog(BLOG_WARNING,
				    "bad value for job delayed start \"%s\"",
				    cp);
			}
			break;
		case 'c':	/* Create time for the job */
			if ((jp->j_ctime = atol(cp)) < 0) {
				blog(BLOG_WARNING,
				    "bad value for job creation time \"%s\"",
				    cp);
				jp->j_ctime = 0;
				jp->j_mtime = 0;
			} else {
				/* Set the mod time to the create time */
				jp->j_mtime = jp->j_ctime;
			}
			break;
		case 'D':	/* working *D*irectory */
			jp->j_wd = estrdup(cp);
			break;
		case 'f':	/* *f*orce the job onto a specific CPU */
			jp->j_mustrun = atoi(cp);
			break;
		case 'R':	/* *R*emote host */
			jp->j_rhost = estrdup(cp);
			break;
		case 'u':	/* numerical *u*id */
#ifdef HAVE_FLAT_NAMES
			if ((id = atoi(cp)) < 0) {
				blog(BLOG_ERR, "can't extract uid from %s",
				    filename);
				return (-1);
			} else
				jp->j_uid = id;
#endif /* FLAT */
			break;
		case 'U':	/* *U*ser name */
			jp->j_user = estrdup(cp);
#ifndef HAVE_FLAT_NAMES
			if ((pw = getpwnam(cp)) == NULL) {
				blog(BLOG_ERR, epent, cp);
				return (-1);
			} else 
				jp->j_uid = pw->pw_uid;
			
			(void) setpwent();
#endif /* FLAT */
			break;
		case 'g':	/* numerical *g*id */
#ifdef HAVE_FLAT_NAMES			
			if ((id = atoi(cp)) < 0) {
				blog(BLOG_ERR, "can't extract gid from %s",
				    filename);
				return (-1);
			} else
				jp->j_gid = id;
#endif /* FLAT */
			break;
		case 'G':	/* *G*roup name */
			jp->j_group = estrdup(cp);
#ifndef HAVE_FLAT_NAMES
			if ((gr = getgrnam(cp)) == NULL) {
				blog(BLOG_ERR, egent, cp);
				return (-1);
			} else
				jp->j_gid = gr->gr_gid;
			
			(void) setgrent();
#endif /* FLAT */
			break;
		case 'j':	/* numerical *j*ob id */
			if ((jp->j_id = atoi(cp)) < 0) {
				blog(BLOG_ERR, "can't extract job id from %s",
				    filename);
				return (-1);
			}
			break;
		case 'O':	/* *O*utput file */
			jp->j_ofile = estrdup(cp);
			break;
		case 'E':	/* *E*rror file */
			jp->j_efile = estrdup(cp);
			break;
		case 'C':	/* Combine - C must come after O or E */
			if (jp->j_ofile && jp->j_efile) 
				blog(BLOG_ERR, "O and E and K specified");
			else if (jp->j_ofile) 
				jp->j_efile = jp->j_ofile;
			else if (jp->j_efile) 
				jp->j_ofile = jp->j_efile;
			else
				blog(BLOG_ERR, "K specified but not O or E");
			break;
		case 'M':	/* *M*ail address */
			jp->j_umaddress = estrdup(cp);
			break;
		case 'm':	/* Mail arguments */
			jp->j_usermail = (Notify) atoi(cp);
			break;
#ifdef HAVE_RLIMIT
		case 'r':	/* Parse user supplied current limits */
			(void) atorlimit(jp->j_rlimit, cp);
			break;
#endif
		case 'p':	/* Attatch to a *p*rocess */
			/*
			 * Eventually we use the stime value to set the
			 * process id should everything go as planned
			 */
			jp->j_stime = atoi(cp);
			jp->j_state |= JOB_ATTACH;
			break;
		case 'w':	/* Write arguments */
			jp->j_userwrite = (Notify) atoi(cp);
			break;
		case 'J':	/* Symbolic *J*ob name */
			jp->j_jname = estrdup(cp);
			break;
#ifdef ndef
		case 'L':	/* *L*ocal - nolonger used */
			if (strcasecmp(cp, yes) == 0)
				jp->j_state |= JOB_LOCAL;
			break;
#endif
		case 'S':	/* Re*S*tart */
			if (strcasecmp(cp, yes) == 0)
				jp->j_restart = TRUE;	
			else
				jp->j_restart = FALSE;
			break;
		case 'W':	/* *W*rite address */
			jp->j_uwaddress = estrdup(cp);
			break;
		case 'Z':	/* Keep cf file as kf file. */
			if (strcasecmp(cp, yes) == 0)
				jp->j_state |= JOB_KEEP;
			break;
		default:	/* Extra stuff is fine */
			break;

		}
	}

	(void) fclose(cfp);

	return 0;
}


static int
job_delay_set(qp, jp)
Queue *	qp;
Job  *	jp;
{

	Queue *	dqp;
	Job *	djp;

	switch(jp->j_delay.d_type) {
		/* No delay type, we're not delayed. */
	case DELAY_NONE:
		break;
	case DELAY_TIME:
		if (jp->j_delay.delay_u.d_time != 0 && jp->j_devent == 0)
			jp->j_devent = event_add((jp->j_delay.delay_u.d_time +
						  stimer) - now,
						 do_delay, (voidp) jp,
						 "delayed time start");
		break;
	case DELAY_FILE:
		if (jp->j_devent == 0)
			jp->j_devent = event_add(ptimer + getrand(-5, 5),
						 do_delay, (voidp) jp,
						 "delayed file start");
		break;
	case DELAY_JOB:
		if ((dqp = getqueuebyname(jp->j_delay.delay_u.d_djob.d_queue)) == NULL) {
			blog(BLOG_ERR, enotq, jp->j_delay.delay_u.d_djob.d_queue);
			return (-1);
		}
		/*
		 * If the job is found in the specified queue,
		 * 	then it is still queued, or running.
		 * else if not there, assume the job has finished,
		 * 	and the current job is now ready to run.
		 */
		ITER(djp, Job, &(dqp->q_jobs)) {
			if (djp->j_id == jp->j_delay.delay_u.d_djob.d_id)
				djp->j_state |= JOB_DELAYER;
		}
		break;
	default:
		blog(BLOG_WARNING, "bad delay type (%d)",
		     jp->j_delay.d_type);
		break;
	}

	return 0;
}

/*
 * Return the usr structure that corresponds to the owner uid or gid of
 * the job in question.
 */
Usr *
getusrbyid(usrs, type, jp)
qelem *	usrs;
int	type;
Job *	jp;
{

	Usr *	up;

	ITER(up, Usr, usrs) {
		if ((type & SORT_UID) && up->u_uid == jp->j_uid) {
			dbg((BLOG_JOBS, "hit on uid: %u", up->u_uid));
			return up;
		} else if ((type & SORT_GID) && up->u_gid == jp->j_gid) {
			dbg((BLOG_JOBS, "hit on gid: %u", up->u_gid));
			return up;
		} else {
			continue;
		}
	}

	return (Usr *) NULL;
}


Usr *
usr_alloc(usrs)
qelem *	usrs;
{

	Usr *	unew;


	/*
	 * Add a new usr slot.
	 */
	dbg((BLOG_ALLOC, "allocating Usr structure"));

	unew = ALLOC(Usr);

	qinfo.i_usrlist++;

	ins_tail_que(&(unew->u_usrs), usrs);

	return unew;
}				


/*
 * Dump the user data as it stands now, using a supplied function.
 */
void
dump_usr(func, lvl, qp)
void	(*func)V((int, char *, ...));
int	lvl;
Queue *	qp;
{

	Usr *	up;
	char *	type;

	if (qp->q_ftype & SORT_UID)
		type = "uid";
	else
		type = "gid";

	ITER(up, Usr, &(qp->q_usrs)) {

		(*func)(lvl,
			"queue: %s %s: %-4d times:%6lds, %6lds submit:%d, queue:%d, complete:%d",
			qp->q_name, type,
			up->u_uid, up->u_cputime, up->u_runtime,
			up->u_submit, up->u_queued, up->u_complete);
	}

	return;
}


/*
 * Return a new job structure either by allocation, or grabbing one off
 * the free list.
 */
static Job *
job_alloc(jobs)
qelem *	jobs;
{

	Job *	jnew;

	if (free_jp > 0) {

		dbg((BLOG_ALLOC, "reusing Job structure"));

		jnew = FIRST(Job, &fjobs);
		rem_que(&jnew->j_jobs);
		free_jp--;
		qinfo.i_fjoblist--;

	} else {

		/*
		 * Make space for a new job entry
		 */
		dbg((BLOG_ALLOC, "allocating Job structure"));

		jnew = ALLOC(Job);

		qinfo.i_joblist++;
	}

	ins_tail_que(&(jnew->j_jobs), jobs);

	return jnew;

}

/*
 * Free() or cubby hole a job structure that we have let go.
 */
void
job_free(jp)
Job *	jp;
{

	/* Here we free all the allocated job member pointers */
	job_free_struct(jp);

	rem_que(&jp->j_jobs);

	/*
	 * Then either save the structure on a free list, or
	 * just free it alltogether.
	 */
	if (free_jp > FREE_JOBS) {

		efree((voidp) jp);

	} else {

		ins_tail_que(&jp->j_jobs, &fjobs);
		free_jp++;
		qinfo.i_fjoblist++;

	}
}
	

/*
 * Free/zero the job structure information.
 */
void
job_free_struct(jp)
Job *	jp;
{

	dbg((BLOG_JOBS, quej, jp->j_queue->q_name, jp->j_id,
	    "freeing structures"));

	efree((voidp) jp->j_file);
	jp->j_file = NULL;

	efree((voidp) jp->j_rhost);
	jp->j_rhost = NULL;

	efree((voidp) jp->j_wd);
	jp->j_wd = NULL;

	efree((voidp) jp->j_ofile);
	jp->j_ofile = NULL;

	efree((voidp) jp->j_efile);
	jp->j_efile = NULL;

	efree((voidp) jp->j_umaddress);
	jp->j_umaddress = NULL;

	efree((voidp) jp->j_uwaddress);
	jp->j_uwaddress = NULL;

	efree((voidp) jp->j_jname);
	jp->j_jname = NULL;

	efree((voidp) jp->j_user);	/* User name of job owner */
	jp->j_user = NULL;

	efree((voidp) jp->j_group);	/* Group name of job owner */
	jp->j_group = NULL;

	jp->j_shost = NULL;
	jp->j_cputime = 0;
	jp->j_rsize = 0;

	/*
	 * Remove a delayed start event if the job is removed.
	 */
	if (jp->j_devent != 0)
		(void) event_remove(jp->j_devent);
	jp->j_devent= 0;

	switch (jp->j_delay.d_type) {
	case DELAY_TIME:
		break;
	case DELAY_FILE:
		efree((voidp) jp->j_delay.delay_u.d_file);
		jp->j_delay.delay_u.d_file = NULL;
		break;
	case DELAY_JOB:
		efree((voidp) jp->j_delay.delay_u.d_djob.d_queue);
		jp->j_delay.delay_u.d_djob.d_queue = NULL;
		efree((voidp) jp->j_delay.delay_u.d_djob.d_host);
		jp->j_delay.delay_u.d_djob.d_host = NULL;
		break;
	}

	jp->j_delay.d_type = 0;
	efree((voidp) jp->j_delay.d_string);
	jp->j_delay.d_string = NULL;

#ifdef DATA_SIZE	
	jp->j_ssize = 0;
	jp->j_dsize = 0;
#endif

	jp->j_id = 0;	
	jp->j_pid = 0;

	jp->j_mustrun = -1;			/* '0' is a valid CPU name */

	jp->j_ndpriority = 0;
	jp->j_priority = 0;

	jp->j_ctime = 0;
	jp->j_mtime = 0;

	/* Remove a delayed (send) transmit event if the job is removed. */
	if (jp->j_stime != 0 && jp->j_sevent != 0)
		(void) event_remove(jp->j_sevent);

	jp->j_stime = 0;
	jp->j_sevent= 0;

	jp->j_state = JOB_NULL;
	jp->j_restart = FALSE;
	jp->j_seen = FALSE;

	jp->j_usr = NULL;
	jp->j_queue = NULL;

	return ;
}


/*
 * Sort the jobs using qsort().
 */
void
job_sort(jobs)
qelem *	jobs;
{
	Job *	jp;
	int	i, count;

	/* Pack the Job pointers into an array */
	count = 0;
	ITER(jp, Job, jobs)
		jarray[count++] = jp;

	/* count == number of Job pointers */

	qsort((voidp) jarray, count, sizeof(Job *), job_comp);

	/* Unpack from the array and recreate the linked list */
	jobs->q_forw = jobs;
	jobs->q_back = jobs;

	for (i = 0; i < count; i++) {
		ins_head_que(&(jarray[i]->j_jobs), jobs);
	}
	return ;
}


/*
 * Sort the job pointers with this function
 */
int
job_comp(i1, i2)
const void *	i1;
const void *	i2;
{

	int	i;
	int	rtn = 0;
	Job **	j1 = (Job **) i1;
	Job **	j2 = (Job **) i2;
	Queue *	qp = (*j1)->j_queue;

	if ((*j1)->j_queue != (*j2)->j_queue)
		blog(BLOG_WARNING,
		     "sanity check fails: j1 & j2 not in same queue");

	/*
	 * First consider the running jobs - they are always at the
	 * head of the list, sorted by the time they started.
	 */
	if (((*j1)->j_state & JOB_RUNNING) && !((*j2)->j_state & JOB_RUNNING))
		return(1);

	if (!((*j1)->j_state & JOB_RUNNING) && ((*j2)->j_state & JOB_RUNNING))
		return (-1);

	if (((*j1)->j_state & JOB_RUNNING) && ((*j2)->j_state & JOB_RUNNING))
		return job_comp_mtime((*j1), (*j2));

	/* else neither job running */

	/*
	 * If one of the jobs has been set with topq, then it goes first.
	 * If both jobs were set, then sort only by mtime, nothing else
	 * takes precedence.
	 */
	if (((*j1)->j_state & JOB_TOPQ) && !((*j2)->j_state & JOB_TOPQ))
		return (1);

	if (!((*j1)->j_state & JOB_TOPQ) && ((*j2)->j_state & JOB_TOPQ))
		return (-1);

	if (((*j1)->j_state & JOB_TOPQ) && ((*j2)->j_state & JOB_TOPQ))
		return job_comp_mtime((*j1), (*j2));

	/* else neither job top queued */
#ifdef fixedsort
	/*
	 * Implicit ordering - round robin, then usage, 
	 * Flip if statements here in the code if different sort order
	 * requested.
	 */
	if (qp->q_ftype & SORT_ROUND) {
		if ((rtn = job_comp_queued((*j1), (*j2))) != 0)
			return rtn;
	}

	if (qp->q_ftype & SORT_USAGE) {
		if ((rtn = job_comp_usage((*j1), (*j2))) != 0)
			return rtn;
	}

	if (qp->q_ftype & SORT_COMPLETE) {
		if ((rtn = job_comp_complete((*j1), (*j2))) != 0)
			return rtn;
	}

	/* Last resort - sort by mtime */
	return job_comp_mtime((*j1), (*j2));

#else

	/* Sort by the order of the functions in the list */
	for (i = 0; i < qp->q_nfunc; i++) {
		if ((rtn = (*(qp->q_func)[i])((*j1), (*j2))) != 0)
			return rtn;
	}

#endif /* fixedsort */

	return rtn;

}


/*
 * Compare mtime of jobs.
 */
int
job_comp_mtime(j1, j2)
Job *	j1, *	j2;
{

	if (j1->j_mtime < j2->j_mtime)
		return (1);
	else if (j1->j_mtime > j2->j_mtime)
		return (-1);
	else
		return (0);
}

/*
 * Compare number of jobs queued.
 */
int
job_comp_queued(j1, j2)
Job *	j1, *	j2;
{

	Usr *u1 = j1->j_usr;
	Usr *u2 = j2->j_usr;

	if (u1 == NULL || u2 == NULL) {
		blog(BLOG_WARNING, "user structure is NULL in comp_queued");
		return (0);
	}

	if (u1->u_queued < u2->u_queued)
		return (1);
	else if (u1->u_queued > u2->u_queued)
		return (-1);
	else
		return (0);
}

/*
 * Compare CPU time usage
 */
int
job_comp_usage(j1, j2)
Job *	j1, *	j2;
{

	int	usage1 = 0;
	int	usage2 = 0;
	Usr *	u1 = j1->j_usr;
	Usr *	u2 = j2->j_usr;
	Queue *	qp = j1->j_queue;


	if (u1 == NULL || u2 == NULL) {
		blog(BLOG_WARNING, "user structure is NULL in comp_usage");
		return (0);
	}

	usage1 = u1->u_cputime + u1->u_runtime;
	usage2 = u2->u_cputime + u2->u_runtime;

	if (qp->q_ulimit > 0) {
		if (usage1 < qp->q_ulimit &&
		    usage2 < qp->q_ulimit) {
			return 0;
		} else if (usage1 < qp->q_ulimit &&
			   usage2 > qp->q_ulimit) {
			return 1;
		} else if (usage1 > qp->q_ulimit &&
			   usage2 < qp->q_ulimit) {
			return -1;
		} else if (usage1 > qp->q_ulimit &&
			   usage2 > qp->q_ulimit) {
			return 0;
		}
	} else {
		if (usage1 < usage2) {
			return 1;
		} else if (usage1 > usage2) {
			return -1;
		} else if (usage1 == usage2) {
			return 0;
		}
	}
	return 0;
}


/*
 * Compare jobs completed.
 */
int
job_comp_complete(j1, j2)
Job *	j1, *	j2;
{

	int	complete1 = 0;
	int	complete2 = 0;
	Usr *	u1 = j1->j_usr;
	Usr *	u2 = j2->j_usr;
	Queue *	qp = j1->j_queue;


	if (u1 == NULL || u2 == NULL) {
		blog(BLOG_WARNING, "user structure is NULL in comp_complete");
		return (0);
	}

	complete1 = u1->u_complete;
	complete2 = u2->u_complete;
	
	if (qp->q_climit > 0) {
		if (complete1 < qp->q_climit &&
		    complete2 < qp->q_climit) {
			return 0;
		} else if (complete1 < qp->q_climit &&
			   complete2 > qp->q_climit) {
			return 1;
		} else if (complete1 > qp->q_climit &&
			   complete2 < qp->q_climit) {
			return -1;
		} else if (complete1 > qp->q_climit &&
			   complete2 > qp->q_climit) {
			return 0;
		}
	} else {
		if (complete1 < complete2) {
			return 1;
		} else if (complete1 > complete2) {
			return -1;
		} else if (complete1 == complete2) {
			return 0;
		}
	}
	return 0;
}


/*
 * Add the job to the queue, and send it SIGSTOP until we are ready to have
 * it run in the queue.
 */
static int
queue_attach_job(qp, jp)
Queue *	qp;
Job *	jp;
{

	Queue *	aqp;
	Job *	ajp;
	static char *	mess =
		"Queue %s: job %d: already running in %s queue\n";


	/* Subject line */
	(void) sprintf(line, quej, qp->q_name, jp->j_id, "not attached");
		
	ITER(aqp, Queue, &queues) {
		ITER(ajp, Job, &(aqp->q_jobs)) {
			if (jp->j_pid == ajp->j_pid) {
				blog(BLOG_STATS, line);
				queue_write_status(qp, line);
				return (-1);
			}
		}
	}


	/*
	 * Set the pid, but do not mark this as a JOB_RUNNING,
	 * effectively the job is still queued.
	 */
	jp->j_pid = jp->j_stime;
	jp->j_stime = 0;
	jp->j_state |= JOB_ATTACHED;

	/* If this queue is stopped, stop the job. */
	if (queue_get_flag(qp, RUNNING) != 0) {
		(void) job_killpg(jp, SIGSTOP);
		(void) sprintf(line, quej, qp->q_name, jp->j_id, "held");

		blog(BLOG_STATS, line);
		queue_write_status(qp, line);
	}

	return 0;
}


