#ifndef lint
static char rcsid[] =
	"$Id: runjob.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */


/*
 * runjob.c - re-analyse and modify the state of execution in the queues.
 *	Start jobs if any are waiting to be started.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>

#include <signal.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/dir.h>
#include <sys/wait.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#ifdef HAVE_SCHEDCTL
#include <sys/prctl.h>
#include <sys/schedctl.h>
#endif

#ifdef HAVE_SYSMP
#include <sys/sysmp.h>
#endif

#include "batch.h"
#include "batchd.h"
#include "event.h"
#include "flags.h"
#include "protocol.h"
#include "pathnames.h"


static int	edup2 P((int, int));
static int	openo P((char *, char *));
static int	queue_start_job P((Queue *, Job *));
static int	queue_start_attached_job P((Queue *, Job *));

static int	job_check_delays P((Queue *, Job *));

static int	queue_renice_jobs P((Queue *qp, int newnice));

static int	set_resource_usage P((Queue *, Job *, int));

extern void	process_handle P((Run *, wait_t *, rusage_t *));

extern void 	do_periodic P((Queue *));
extern void 	do_delay P((Job *));

static char *	astart = "start %-6.6s uid:%-5d job:%-5d pid:%-5d";
static char *	mstart = "Queue %s: job %d (%s) started for %s.";


/*
 * Determine the state of the queue for execution.  Attempt to start jobs
 * if deemed reasonable.
 *
 * Return the following:
 * STATUS_ERROR - 	an error occured in the normal queueing routine.
 * STATUS_DISABLED -	the queue is or became disabled. 
 * STATUS_NOCHANGE -	nothing happened.
 * STATUS_ENABLED -	the queue became enabled or a job was queued.
 * STATUS_PERSIST -	there was a message put into the queue status file.
 */
int
queue_run_jobs(qp, force)
Queue *	qp;
int	force;
{

	int	status = STATUS_NOCHANGE;
	int	rtn = STATUS_NOCHANGE;
	Bool	change = FALSE;
	Job *	jp;				/* Current job */
	Usr *	up;				/* Current user */

	/*
	 * Change is TRUE if a queue is to be disabled in some way.
	 * Logic: queuestate && whether the user wants to examine this
	 * change of state.
	 */
	change = (queue_get_lock(qp, STOPPED) == 0);
	if ((rtn =  queue_change_state(qp, change, STOPPED, force))
	    != STATUS_ENABLED) {
		if (change) {
			if (queue_get_lock(qp, DISABLED) == 0)
				return (STATUS_PERSIST);
			else
				return (STATUS_DISABLED);
		}
	} else {
		status = rtn;
	}
	
	change = (!qp->q_timein &&
		  (qp->q_userrc ? (qp->q_uctl & TIMEOUT) : TRUE));
	if ((rtn = queue_change_state(qp, change, TIMEOUT, force))
	    != STATUS_NOCHANGE)
		status = rtn;

	change = (((int) (loadave[LOAD_5MIN] + 0.5) > qp->q_loadstop) &&
		  (qp->q_userrc ? (qp->q_uctl & SUSPEND) : TRUE));
	if ((rtn = queue_change_state(qp, change, SUSPEND, force))
	    != STATUS_NOCHANGE)
		status = rtn;

	change = (!qp->q_inactive && 
		  (qp->q_userrc ? (qp->q_uctl & INTERRUPT) : TRUE));
	if ((rtn = queue_change_state(qp, change, INTERRUPT, force))
	    != STATUS_NOCHANGE)
		status = rtn;

	/*
	 * If the queue is draining, don't queue any more processes,
	 * just maintain those that are running.
	 */
	change = (queue_get_lock(qp, DRAINING) == 0);
	if ((rtn =  queue_change_state(qp, change, DRAINING, force))
	    != STATUS_ENABLED) {
		if (change) {
			return (STATUS_DISABLED);
		}
	} else {
		status = rtn;
	}
	
	/* If RUNNING is not set, return from here. */
	if (queue_get_flag(qp, RUNNING) != 0)
		return (status);

	blog(BLOG_TRACE, "%s: now running jobs (load: %d)",
	    qp->q_name, (int) (loadave[LOAD_5MIN] + 0.5));


	ITER(jp, Job, &(qp->q_jobs))
		jp->j_seen = FALSE;

	/*
	 * For each active, unprocessed job in the queue...
	 */
	ITER(jp, Job, &(qp->q_jobs)) {

		/* If the job is running, skip it. */
		if (jp->j_state & JOB_RUNNING)
			continue;

		/*
		 * Check on delayed jobs.
		 */
		if (job_check_delays(qp, jp) < 0)
			continue;
		
		/*
		 * If it's to be attached or run, and there is room,
		 * start the job.
		 */
		if ((jp->j_state & JOB_ATTACH) &&
		    qp->q_nexec < qp->q_maxexec) {
			dbg((BLOG_JOBS, "attached job: %d", jp->j_id));
			if (queue_start_attached_job(qp, jp) < 0)
				continue;
			else
				status = STATUS_PERSIST;
		} else if ((jp->j_state & JOB_RUN) &&
		    qp->q_nexec < qp->q_maxexec) {
			dbg((BLOG_JOBS, "run job: %d", jp->j_id));
			if (queue_start_job(qp, jp) < 0)
				continue;
			else
				status = STATUS_PERSIST;
		} else
			continue;

	}

	return status;
}


/*
 * Set the state of the queue and signal jobs of that queue as a side effect
 * Return STATUS_DISABLED/ENABLED or NOCHANGE, if so.
 */
int
queue_change_state(qp, action, flag, force)
Queue *	qp;
int	action;
State	flag;
int	force;
{

	int	i;
	int	status = STATUS_NOCHANGE;
	char *	up = NULL;
	char *	down = NULL;
	State	go, check;

	if (flag & ANYRUN) {
		go = RUNNING;
		check = ANYRUN;
	} else if (flag & ANYSUB) {
		go = QUEUING;
		check = ANYSUB;
	} else if (flag & DRAINING) {
		go = 0;
		check = 0;
	} else
		blog(BLOG_WARNING, "undefined flag type (%lo)", flag);

	/* Fit the names to use for this flag */
	for (i = 0; i < sizeof(namestab)/sizeof(struct names_t); i++)
		if (namestab[i].n_flag == flag) {
			up = namestab[i].n_up;
			down = namestab[i].n_down;
			break;
		}

	if (up == NULL)
		blog(BLOG_WARNING, "couldn't find state names");

	/* 
	 * If the action is to add the given flag as a restriction:
	 * 1) See that it is not already a restriction, then add it
	 * 2) If this is the first restriction, send SIGSTOP to all
	 * 	the jobs in that queue.
	 */

	if (action) {

		/*
		 * If not set, then set and mark fresh set logs,
		 * else remind us that it is set in the logs
		 */
		if (queue_get_flag(qp, flag) != 0) {
			queue_set_flag(qp, flag);
			queue_set_lock(qp, flag);
			blog(BLOG_TRACE, "%s: freshly set %s",
			    qp->q_name, down);
			status = STATUS_DISABLED;
		} else
			blog(BLOG_TRACE, ques, qp->q_name, down);

		if (go && queue_get_flag(qp, go) == 0) {
			queue_clr_flag(qp, go);
			if (check == ANYRUN)
				queue_signal_jobs(qp, SIGSTOP, down);
		} 
		/* else, were already disabled */

	/*
	 * If the action is to remove the current restriction of the
	 * flag argument:
	 * 1) Check to see that the flag is actually set, then remove it.
	 * 2) If that is the last pending condition in that queue, send
	 * 	SIGCONT to the jobs in that queue.
	 * 3) Return the status 
	 */
	} else /* if (!action) */ {

		/*
		 * If flag is set, clear it and mark fresh in the logs,
		 * else remind us that it is not set in the logs.
		 */
		if (queue_get_flag(qp, flag) == 0) {
			queue_clr_flag(qp, flag);
			blog(BLOG_TRACE, "%s: freshly set %s",
			    qp->q_name, up);
			status = STATUS_ENABLED;
		} else
			blog(BLOG_TRACE, ques, qp->q_name, up);

		/*
		 * If no running flags are set, set running and continue
		 * the jobs.
		 */
		if (check && queue_get_flag(qp, check) != 0) {
			queue_set_flag(qp, go);
			queue_set_lock(qp, go);
			if (check == ANYRUN)
				queue_signal_jobs(qp, SIGCONT, up);
		}
		/* else, something is still blocking us. */
	}

	return (status);
}	


/*
 * Find the CPU time used and possibly terminate each job.
 * Return the number of jobs found and examined;
 */
int
queue_limit_jobs(qp)
Queue *	qp;
{


	Job *	jp;
	Usr *	up;
	int	found = 0;

	if (job_init_usage() < 0) {
		blog(BLOG_WARNING, ques, qp->q_name,
		     "proc structure not read");
		return -1;
	} else {
		dbg((BLOG_PROC, ques, qp->q_name,
		     "proc structure read"));
	}

	/* Clear the runtime CPU counter for each user */
	ITER(up, Usr, &(qp->q_usrs)) {
		up->u_runtime = 0;
	}

	ITER(jp, Job, &(qp->q_jobs)) {

		/* If job is not running, don't check CPU time */
		if (!(jp->j_state & JOB_RUNNING))
			continue;

		if (job_get_usage(jp) != RTN_OK) {
			blog((BLOG_PROC | BLOG_TRACE),
			    quej, qp->q_name, jp->j_id,
			     "job_get_usage(): failed");
			continue;
		}

		found++;

		if (job_check_usage(jp) == RTN_REPORT) {
			blog((BLOG_PROC | BLOG_TRACE), 
			     quej, qp->q_name, jp->j_id, "was killed");
		} else {
			/*
			 * If the job was found, and not
			 * killed, do our internal accounting.
			 */
			if ((up = getusrbyid(&(qp->q_usrs), SORT_UID, jp)) == NULL) {
				blog(BLOG_WARNING, "%s: cannot find uid: %u",
				     qp->q_name, jp->j_uid);
			} else {
				up->u_runtime += jp->j_cputime;
			}
		}
	}

	return found;
}


/*
 * Recalculate the nice values for each queue based on the number
 * of jobs in the system, and the span of nice values for the queue.
 * Return the new nice used for this queue;
 */
int
queue_prioritize_jobs(qp)
Queue *	qp;
{

	int	nexec = 0, maxexec = 0;
	int	low, high, span;
	int	newnice = NICE_MIN;
	double	pexec;
	Job *	jp;
	Queue *	aqp;				/* Queue for newnice sum */


	if (qp->q_priority[NICE_LOW] != qp->q_priority[NICE_HIGH]) {

		nexec = 0; maxexec = 0;	pexec = 0;

		/* Reset the nexec and maxexec if using separate count */
		if (qp->q_ptype & NICE_SEPARATE) {
			nexec = qp->q_nexec;
			maxexec = qp->q_maxexec;
		/* Sumup nexec and maxexec if using ganged count */
		} else if (qp->q_ptype & NICE_GANG) {
			ITER(aqp, Queue, &queues) {
				nexec += aqp->q_nexec;
				maxexec += aqp->q_maxexec;
			}

		}

		if (maxexec != 0)
			pexec = (double) nexec/maxexec;

		/* Fold in current and max load, and recalcuate percentage */
		if (qp->q_ptype & NICE_LOAD) {
			nexec += (int) (loadave[LOAD_5MIN] + 0.5);
			maxexec += qp->q_loadstop;
			if (maxexec != 0)
				pexec = (double) nexec/maxexec;
		}

		/* If more than one job running, jump to low priority */
		if (qp->q_ptype & NICE_JUMP) {
			if (nexec > 1)
				newnice = qp->q_priority[NICE_LOW];
			else 
				newnice = qp->q_priority[NICE_HIGH];

		/* If more than half maximum percent, step to low priority */
		} else if (qp->q_ptype & NICE_STEP) {
			if (pexec >= 0.5)
				newnice = qp->q_priority[NICE_LOW];
			else 
				newnice = qp->q_priority[NICE_HIGH];

		/* else interpolate (by ints) to the appropriate priority */
		} else if (qp->q_ptype & NICE_INTERP) {

			/* lowest priority (closest to +20/40) */
			low = qp->q_priority[NICE_LOW];

			/* closer to -20/0 */
			high = qp->q_priority[NICE_HIGH];

			span = low - high;

			newnice = (int) ((span * pexec) + 0.5) + high;
		} else {
			blog(BLOG_WARNING, "bad reschedule parameters");
		}

		dbg((BLOG_RESOURCE,
		     "%s: pexec: %.2f, span: %d, setpriority to %d",
		     qp->q_name, pexec, span, newnice));

		(void) queue_renice_jobs(qp, newnice);
	}

	return newnice;
}


/*
 * Renice all jobs in a queue to new nice value
 * Return the number of jobs reniced;
 */
static int 
queue_renice_jobs(qp, newnice)
Queue *	qp;
int	newnice;
{

	int	cfd;
	int	found = 0;
	Job *	jp;


	now = time((time_t *)0);

	ITER(jp, Job, &(qp->q_jobs)) {

		/* Nice value is correct - don't reset */
		if (jp->j_priority == newnice)
			continue;

		/* setpriority might return valid neg number so check errno */
		errno = 0;
		if (setpriority(PRIO_PGRP, jp->j_pid, newnice) == -1 &&
		    errno != 0) {
			blog(BLOG_ERR, quejr, qp->q_name, jp->j_id,
			     "renice jobs setpriority()");
			continue;
		} else {
			blog(BLOG_TRACE, "%s: %d setpriority to %d",
			     qp->q_name, jp->j_id, newnice);

			jp->j_priority = newnice;
			found++;
		}

		/* Update the control file with priority change */
		if ((cfd = open(jp->j_file, (O_RDWR|O_APPEND))) < 0) {
			blog(BLOG_WARNING, eopen, jp->j_file);
		} else {
			(void) sprintf(line, "%c%s %d %d %ld\n",
				       'N', lhost, jp->j_pid, newnice, now);
			(void) ewrite(cfd, line, strlen(line), "startjob()");
			(void) close(cfd);
		}

	}

	return found;
}


/*
 * Reap attached jobs in this queue that have completed.
 * Return tthe number of jobs reaped.
 */
int
queue_attach_reap(qp)
Queue *	qp;
{

	int	found = 0;		/* number of jobs reaped */
	Job *	jp;
	Run *	rp;


	/*
	 * Check to see that the attached jobs are still running...
	 */
	ITER(jp, Job, &(qp->q_jobs)) {

		/* If job doesn't have a pid, job is not running */
		if (!jp->j_pid)
			continue;
		
		/* if kill 0 succeeds, job exists */
		if (kill(jp->j_pid, 0) >= 0)
			continue;

		/* if job not ATTACH, skip */
		if (!(jp->j_state & JOB_ATTACH))
			continue;

		/*
		 * These are now jobs that have a pid, don't exist and
		 * are maked ATTACH.
		 */
		ITER(rp, Run, &runs)
			if (rp->r_pid == jp->j_pid)
				break;
	
		/*
		 * This should never happen, provided SIGCHLD is properly
		 * held from before the fork until after the process is
		 * put in the run table.
		 */
		if (rp == HEAD(Run, &runs)) {
			blog(BLOG_ERR, "%s: process %d *not* in proc table?",
			    "queue_attach_reap", jp->j_pid);
			continue;
		}

		nkids--;
	
		process_handle(rp, (wait_t *) 0, (rusage_t *) 0);
		found++;

		jp = FIRST(Job, &(qp->q_jobs));
	}

	return found;
}

static int
job_check_delays(qp, jp)
Queue *	qp;
Job *	jp;
{

	int	found = 0;
	Queue *	dqp;				/* Queue to check on delay */
	Job *	djp;				/* Job to check on delay */


	switch (jp->j_delay.d_type) {
	case DELAY_NONE:
		/* No delay type, we're not delayed. */
		break;
	case DELAY_TIME:
		/* If we haven't passed the start time, continue. */
		if (jp->j_delay.delay_u.d_time > 0 &&
		    (jp->j_state & JOB_DELAY)) {
			if (jp->j_delay.delay_u.d_time < time((long *) 0)) {
				jp->j_state &= ~JOB_DELAY;
				dbg((BLOG_JOBS,
				     "job %d is now time startable",
				     jp->j_id));
			} else {
				dbg((BLOG_JOBS, "delay job: %d", jp->j_id));
				return (-1);	/* Not time yet */
			}
		}
		break;
	case DELAY_FILE:
		/* If the file to delay on is not here, continue. */
		if (jp->j_delay.delay_u.d_file &&
		    (jp->j_state & JOB_DELAY)) {
			if (*(jp->j_delay.delay_u.d_file) != '/') {
				(void) strcpy(line, jp->j_wd);
				(void) strcat(line, "/");
				(void) strcat(line, jp->j_delay.delay_u.d_file);
			} else {
				(void) strcpy(line, jp->j_delay.delay_u.d_file);
			}
			if (access(line, F_OK) == 0) {
				jp->j_state &= ~JOB_DELAY;
				dbg((BLOG_JOBS,
				     "job %d is now file startable",
				     jp->j_id));
			} else {
				dbg((BLOG_JOBS, "delay job: %d", jp->j_id));
				/* Set up for another delay? */
				if (jp->j_devent == 0) {
					jp->j_devent = event_add(ptimer + getrand(-5, 5),
								 do_delay, (voidp) jp,
								 "delayed file start");
				}
				return (-1);	/* Not time yet */
			}
		} 
		break;
	case DELAY_JOB:
		if ((dqp = getqueuebyname(jp->j_delay.delay_u.d_djob.d_queue)) == NULL) {
			blog(BLOG_ERR, enotq, jp->j_delay.delay_u.d_djob.d_queue);
			return (-1);
		}
		/*
		 * If the job is found in the specified queue,
		 * 	then it is still queued, or running.
		 * else if not there, assume the job has finished,
		 * 	and the current job is now ready to run.
		 */
		found = 0;

		ITER(djp, Job, &(dqp->q_jobs)) {
			if (djp->j_id == jp->j_delay.delay_u.d_djob.d_id)
				found = 1;
		}
		if (!found) {
			jp->j_state &= ~JOB_DELAY;
			dbg((BLOG_JOBS,
			     "job %d is now job startable",
			     jp->j_id));
		} else {
			dbg((BLOG_JOBS,
			     "delay job: %d", jp->j_id));
			return (-1);	/* Not time yet */
		}
		break;
	default:
		blog(BLOG_WARNING, "bad delay type (%d)",
		     jp->j_delay.d_type);
		break;
	}

	return 0;
}


/*
 *  Fork and start a batch job.
 */
static int
queue_start_job(qp, jp)
Queue *	qp;
Job *	jp;
{

	int	i, pid;
	int	fd[3];
	time_t	current;
	int	cfd = 0;

	/* This should never happen */
	if (jp == NULL || qp == NULL)
		return (-1);

	current = time((long *)0);

	fd[0] = -1;				/* "of" file */
	fd[1] = -1;				/* stdout file */
	fd[2] = -1;				/* stderr file */

	(void) sighold(SIGCHLD);		/* until in parent table */

	switch (pid = fork()) {
	case -1:	/* ERROR */
		(void) sigrelse(SIGCHLD);
		blog(BLOG_ERR, efork, "start job");
		return (RTN_ERROR);
		break;
	case 0:		/* CHILD -> start batch job */
		/* for symmetry and completeness */
		(void) sigrelse(SIGCHLD);
		break;
	default:	/* PARENT -> add process, set messages, return */
		/* Increment the count of jobs running in this queue */
		qp->q_nexec++;
			
		/* Increment total number of jobs */
		njobs++;

		jp->j_pid = pid;

		/* Mark the job as running */
		jp->j_state |= JOB_RUNNING;

		/* Add the process to the table	 */
		process_add(pid, qp, jp, "start job");

		/* Job is safely in table now */
		(void) sigrelse(SIGCHLD);

		/* Re-Set the start time */
		jp->j_stime = current;

		/* Update the control file with job start mark  */
		if ((cfd = open(jp->j_file, (O_RDWR|O_APPEND))) < 0) {
			blog(BLOG_WARNING, eopen, jp->j_file);
		} else {
			(void) sprintf(line, "%c%s %d %d %ld\n",
				       'P', lhost, pid, 0, current);
			(void) ewrite(cfd, line, strlen(line), "startjob()");
			(void) close(cfd);
		}

		/* Subject line */
		(void) sprintf(line, quej, qp->q_name, jp->j_id, "started");

		blog(BLOG_STATS, line);
		queue_write_status(qp, line);

		if (jp->j_usermail & NOTIFY_START)
			send_mail(jp->j_umaddress, line,
				  mstart, qp->q_name, jp->j_id, jp->j_jname,
				  jp->j_user);

		if (qp->q_supmail & NOTIFY_START)
			send_mail(qp->q_smaddress, line,
				  mstart, qp->q_name, jp->j_id, jp->j_jname,
				  jp->j_user);

		if (jp->j_userwrite & NOTIFY_START)
			send_write(jp->j_uwaddress,
				   mstart, qp->q_name, jp->j_id, jp->j_jname,
				   jp->j_user);

		if (qp->q_supwrite & NOTIFY_START)
			send_write(qp->q_swaddress,
				   mstart, qp->q_name, jp->j_id, jp->j_jname,
				   jp->j_user);

		if (aflag > 0)
			account(qp, astart, qp->q_name, jp->j_uid,
				jp->j_id, pid);

		qp->q_stats.s_started++;

		/* We're the parent, we go back to our usual business. */
		return (0);
		break;
	}

	/*
	 * This is the child:
	 *	1) Open a general purpose output file.
	 *	2) Make the appropriate dup2's.
	 *	3) Close up the other files.
	 *	4) setpgrp()
	 *	5) Set process limits.
	 *	6) setgid(), setuid()
	 *	7) Open and dup2 user specified files.
	 *	8) execl();
	 */


	/* I/O redirection stuff to mail it back */
	jp->j_file[0] = 'o';	/* Output to `of' file name */

	/*
	 * We must always make an output file even if it is not used, so
	 * that a restart can be detected.  If we don't use it, it will
	 * be closed, a few lines down.
	 */
	/*
	 * Right now, if the output file is on a NFS mounted file system
	 * we loose the ability to write it when we setuid(). 
	 * This seems to be a bug in NFS, it works if the file system
	 * is local.
	 */
	if ((fd[0] = creat(jp->j_file, OFMOD)) < 0) {
		blog(BLOG_FATAL, eopen, jp->j_file);
		exit(RTN_ERROR);
	}

	/*
	 * If the pointer for [stdout|stderr] is null, dup2 the descriptor 
	 * to the newly opened file.
	 */
	if (jp->j_ofile == NULL)
		(void) edup2(fd[0], 1);

	if (jp->j_efile == NULL)
		(void) edup2(fd[0], 2);

	/* Close all open files - except message/log file */
	for (i = 3; i < NOFILE; i++) {
		if (logfp != NULL && i == fileno(logfp)) {
			continue;
		} else {
			(void) close(i);
		}
	}

	errno = 0;

	/* Close stdin and reopen it do /dev/null */
	(void) close(0);
	(void) open(_PATH_DEVNULL, 0);

	/*
	 * For each forked job, reset the process group so that any forks
	 * started *under* the job will have a common but indpendent process
	 * group.  This can then be used to signal the job as a process group.
	 */
#ifdef HAVE_SETSID
	if (setsid() < 0)
#else
#ifdef HAVE_BSDPGRP
	if (setpgrp(0, 0) < 0)
#else
	if (setpgrp() < 0)
#endif
#endif
	{
		blog(BLOG_FATAL, "setpgrp failed in runjob");
		exit(RTN_ERROR);
	}

	/* We only care about the data file now */
	jp->j_file[0] = 'd';

	/* This is a forked job - exit on error */
	set_resource_usage(qp, jp, 1);	

#ifdef HAVE_SYSMP
	if (jp->j_mustrun >= 0)
		if (sysmp(MP_MUSTRUN, jp->j_mustrun) < 0) {
			blog(BLOG_FATAL, quejr, qp->q_name, jp->j_id, "sysmp()");
			exit(RTN_ERROR);
		}

#endif /* HAVE_SYSMP */

#ifdef HAVE_GETGROUPS
	/* Only possible error is uid != 0 */
	(void) initgroups(jp->j_user, (GRPTYPE) jp->j_gid);
#endif

	if (setgid(jp->j_gid) < 0) {
		blog(BLOG_FATAL, quejr, qp->q_name, jp->j_id, "setgid()");
		exit(RTN_ERROR);
	}

	if (setuid(jp->j_uid) < 0) {
		blog(BLOG_FATAL, quej, qp->q_name, jp->j_id, "setuid()");
		exit(RTN_ERROR);
	}

	/*
	 * If the ofile == efile then open one and associate both stdout
	 * and stderr with it, else try opening a file for each if the
	 * name is valid.
	 */
	if ((jp->j_ofile == jp->j_efile) && jp->j_ofile != NULL) {

		fd[1] = openo(jp->j_wd, jp->j_ofile);
		(void) edup2(fd[1], 1);
		(void) edup2(fd[1], 2);

	} else {

		if (jp->j_ofile != NULL) {
			fd[1] = openo(jp->j_wd, jp->j_ofile);
			(void) edup2(fd[1], 1);
		}

		if (jp->j_efile != NULL) { 
			fd[2] = openo(jp->j_wd, jp->j_efile);
			(void) edup2(fd[2], 2);
		}
	}

	/* Reset the signals */
	for (i=1; i <= NSIG; i++)
		(void) signal(i, SIG_DFL);

	errno = 0;

	/* Start the job */
	(void) execl(jp->j_file, jp->j_file, (char *) NULL);

	blog(BLOG_FATAL, "%s: cannot execl() %s: %m", qp->q_name, jp->j_file);

	exit(RTN_ERROR);

	/*NOTREACHED*/
}


/*
 * Create a file.  If the name begins with a '/' take it as a full path,
 * otherwise prepend the current directory of the job directory.
 * Return the file handle.
 */
static int
openo(dir, name)
char *	dir;
char *	name;
{

	int	handle = -1;


	if (*name != '/') {
		(void) strcpy(line, dir);
		(void) strcat(line, "/");
		(void) strcat(line, name);
	} else {
		(void) strcpy(line, name);
	}

	/* Append to the file if it exists, else, create anew. */
	if ((handle = open(line, (O_RDWR|O_APPEND|O_CREAT), OFMOD)) < 0)
		blog(BLOG_FATAL, eopen, line);

	return handle;
}

static int
edup2(from, to)
int	from, to;
{

	int	handle;

	if ((handle = dup2(from, to)) < 0)
		blog(BLOG_FATAL, edup, to);

	return handle;
}



static int
queue_start_attached_job(qp, jp)
Queue *	qp;
Job *	jp;
{

	int	cfd;
	time_t	current;

	/* Set stime to now. */
	jp->j_stime = time((time_t *) 0);
	jp->j_state |= JOB_RUNNING;

	njobs++;

	(void) sighold(SIGCHLD);		/* until in parent table */

	/* Add the process to the table */
	process_add(jp->j_pid, qp, jp, "attached job");

	(void) sigrelse(SIGCHLD);

	/* Update the control file with job start mark */
	if ((cfd = open(jp->j_file, (O_RDWR|O_APPEND))) < 0) {
		blog(BLOG_WARNING, eopen, jp->j_file);
	} else {
		(void) sprintf(line, "%c%s %d %d %ld\n",
			       'P', lhost, jp->j_pid, 1, current);
		(void) ewrite(cfd, line, strlen(line), "startjob()");
		(void) close(cfd);
	}

	qp->q_nexec++;

	/* Start the job if it was stopped */
	(void) job_killpg(jp, SIGCONT);

	/* Subject line */
	(void) sprintf(line, quej, qp->q_name, jp->j_id, "attached");

	blog(BLOG_STATS, line);
	queue_write_status(qp, line);

	if (jp->j_usermail & NOTIFY_START)
		send_mail(jp->j_umaddress, line,
			  mstart, qp->q_name, jp->j_id, jp->j_jname,
			  jp->j_user);

	if (qp->q_supmail & NOTIFY_START)
		send_mail(qp->q_smaddress, line,
			  mstart, qp->q_name, jp->j_id, jp->j_jname,
			  jp->j_user);

	if (jp->j_userwrite & NOTIFY_START)
		send_write(jp->j_uwaddress,
			   mstart, qp->q_name, jp->j_id, jp->j_jname,
			   jp->j_user);

	if (qp->q_supwrite & NOTIFY_START)
		send_write(qp->q_swaddress,
			   mstart, qp->q_name, jp->j_id, jp->j_jname,
			   jp->j_user);

	if (aflag > 0)
		account(qp, astart, qp->q_name,
			jp->j_uid, jp->j_id, jp->j_pid);

	qp->q_stats.s_started++;

	/* Not a forked job - don't punt */
	set_resource_usage(qp, jp, 0);

	return 0;
}


static int
set_resource_usage(qp, jp, forked)
Queue *	qp;
Job *	jp;
int	forked;
{

#ifdef HAVE_RLIMIT


	int	i;
	struct rlimit rl;

	if (forked)
	for (i = 0; i < RLIM_NLIMITS; i++) {

		/*
		 * If watching cumulative CPU time, leave the rlimit
		 * set to the default, otherwise SIGXCPU might be sent to
		 * the process when not warrented.
		 */
		if (qp->q_cumetime == TRUE && i == RLIMIT_CPU) {

			dbg((BLOG_JOBS, "rlimit(CPU) not set: %d %d %d", i,
			     jp->j_rlimit[i].rlim_cur,
			     jp->j_rlimit[i].rlim_max));

			continue;
		}

		dbg((BLOG_JOBS, "rlimit before set: %d %d %d",
		    i, jp->j_rlimit[i].rlim_cur, jp->j_rlimit[i].rlim_max));

		rl.rlim_cur = jp->j_rlimit[i].rlim_cur;
		rl.rlim_max = jp->j_rlimit[i].rlim_max;

		if (rl.rlim_cur >= 0 && rl.rlim_max >= 0)
			if (setrlimit(i, &rl) < 0) {
				blog(BLOG_FATAL,
				    "queue %s: setrlimit(%d) error: (%m)",
				     qp->q_name, i);
				if (forked)
					exit(RTN_ERROR);
			}

		if (getrlimit(i, &rl) < 0) {
			blog(BLOG_FATAL,
			    "queue %s: setrlimit(%d) error: (%m)",
			     qp->q_name, i);
			if (forked)
				exit(RTN_ERROR);

		} else {

			dbg((BLOG_JOBS, "rlimit after  set: %d %d %d",
			    i, rl.rlim_cur, rl.rlim_max));
		}
	}

#endif /* HAVE_RLIMIT */

#ifdef HAVE_SCHEDCTL

	/*
	 * If we have control over scheduling, set the non-degrading
	 * priority to the requested value.
	 */
	if (jp->j_ndpriority >= 0) {
		if (schedctl(NDPRI, jp->j_pid, jp->j_ndpriority) < 0) {
			blog(BLOG_FATAL, quejr, qp->q_name, jp->j_id,
			    "schedctl()");
			if (forked)
				exit(RTN_ERROR);
		}
	}

#endif /* HAVE_SCHEDCTL */

	/* Set the "nice" value if >= 0 */
	if (jp->j_priority >= 0) {
		/* setpriority might return valid neg number so check errno */
		errno = 0;
		if (setpriority(PRIO_PGRP, jp->j_pid, jp->j_priority) == -1 &&
		    errno != 0) {
			blog(BLOG_FATAL,
			     quejr, qp->q_name, jp->j_id, "resource setpriority()");
			if (forked)
				exit(RTN_ERROR);
		}
	}

	return 0;
}
