#ifndef lint
static char rcsid[] =
	"$Id: sendjob.c,v 4.0 1993/09/17 00:41:39 srp Exp srp $";
#endif

/*
 *
 * Author: Scott R. Presnell
 *
 * Copyright (C) 1989, 1990, The Regents of the University of California
 *                     All Rights Reserved
 *
 * THE REGENTS OF THE UNIVERSITY OF CALIFORNIA MAKE NO REPRESENTATION OR
 * WARRANTIES WITH RESPECT TO THE CONTENTS HEREOF AND SPECIFICALLY DISCLAIM
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
 * PURPOSE.
 *
 * Further, the Regents of the University of California reserve the right to
 * revise this software and/or documentation and to make changes from time
 * to time in the content hereof without obligation of the Regents of the
 * University of California to notify any person of such revision or change.
 *
 */


/*
 * sendjob.c - Send a job to a remote host running batchd
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <errno.h>
#include <signal.h>
#include <netdb.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "config.h"

#include H_STRING

#include "batch.h"
#include "batchd.h"
#include "protocol.h"
#include "flags.h"


static int pfd = -1;		/* Global file descriptor for sending jobs. */

static SIGRTN	quit P((int));
static SIGRTN	die P((int));

static int	send_get_response P((void));
static int	send_check_response P((char *));
static int	job_send_file P((char *, int));
static void	send_abort P((void));
static void	send_done P((void));
static void	doexit P((int));

/*
 * Send the daemon control file (cf) and data files (df) over to the remote
 * batchd host.  Return:
 *	-1 if a non-recoverable error occured,
 * 	 1 if a recoverable error and
 *	 0 if all is well.
 */
int
job_send(jp)
Job *	jp;
{

	Queue *	qp = jp->j_queue;
	char  *	queue;
	int	tries = 0;
	int	resp = -1, pid, rtn;
	unsigned int n;
	time_t	current;
	int	cfd = 0;


	current = time((long *) 0);

	if (jp->j_state & JOB_AGAIN) {
		if (jp->j_stime <= current) {
			jp->j_state &= ~JOB_AGAIN;
			jp->j_stime = 0;
		} else {
			return(RTN_RETRY);
		}
	}

	/*
	 * If rq is set, use it for the queue name
	 */
	if (qp->q_rname != NULL)
		queue = qp->q_rname;
	else
		queue = qp->q_name;

	(void) sighold(SIGCHLD);		/* until in parent table */

	switch (pid = fork()) {
	case -1:	/* ERROR */
		(void) sigrelse(SIGCHLD);
		blog(BLOG_ERR, efork, "send job");
		return (RTN_ERROR);
		break;
	case 0:		/* CHILD -> transfer job */
		/*
		 * Indicate a job for transmission
		 */
		jp->j_state |= JOB_SEND;
		jp->j_state &= ~JOB_RUN;
		/* for symmetry and completeness */
		(void) sigrelse(SIGCHLD);
		break;
	default:	/* PARENT - add process, send messages, return */
		/*
		 * Indicate a job for transmission
		 */
		jp->j_state |= JOB_SEND;
		jp->j_state &= ~JOB_RUN;
		
		jp->j_pid = pid;

		/*
		 * Add the process to the table
		 */
		process_add(pid, qp, jp, "send job");
		
		/*
		 * Job is safely in table now
		 */
		(void) sigrelse(SIGCHLD);

		/*
		 * Update the control file with an audit trail of the transfer
		 */
		if ((cfd = open(jp->j_file, (O_RDWR|O_APPEND))) < 0) {
			blog(BLOG_WARNING, eopen, jp->j_file);
		} else {
			(void) sprintf(line, "%c%s %d %d %ld\n",
				       'T', jp->j_shost, pid, 0, current);
			(void) ewrite(cfd, line, strlen(line), "sendjob()");
			(void) close(cfd);
		}

		(void) sprintf(line, "%s: sending job %d to %s@%s",
			       qp->q_name, jp->j_id, queue, jp->j_shost);

		blog(BLOG_STATS, line);
		queue_write_status(qp, line);

		/*
		 * User notification is done in the reaper, only after the
		 * the job as trasnferred with success.
		 */

		qp->q_stats.s_transfered++;

		/*
		 * We're the parent, we go back to our usual business
		 */
		return (RTN_OK);
	}	
	
	/*
	 * SIGKILL will return signal == 9 in the reaper
	 */
	(void) signal(SIGHUP, quit);
	(void) signal(SIGINT, die);
	(void) signal(SIGIOT, SIG_IGN);
	(void) signal(SIGALRM, SIG_IGN);
	(void) signal(SIGPIPE, SIG_IGN);
	(void) signal(SIGUSR1, SIG_IGN);
	(void) signal(SIGUSR2, SIG_IGN);
	(void) signal(SIGTERM, die);

	/*
	 * Set the process group to the process id (job id) so that it
	 * may be killed with a killpg style call to kill.
	 */
#ifdef HAVE_SETSID
	(void) setsid();
#else
#ifdef HAVE_BSDPGRP
	(void) setpgrp(0, 0);
#else
	(void) setpgrp();
#endif
#endif

	/*
	 * Any way that getsocket fails, it returns (-1).
	 */
	if ((pfd = getsocket(jp->j_shost, service, tcp,
			     (u_long *) 0, (u_short *) 0)) < 0) {

		/*
		 * h_errno will be set by gethostbyname() if there 
		 * is a problem.
		 */
		if (h_errno == HOST_NOT_FOUND || h_errno == NO_DATA) {

			dbg((BLOG_XFER, eaddr, jp->j_shost));
			queue_write_status(qp, eaddr, jp->j_shost);

			doexit(RTN_ERROR);
		} else {
			(void) sprintf(line, "%s: cannot connect to host %s",
				       qp->q_name, jp->j_shost);

			dbg((BLOG_XFER, line));
			queue_write_status(qp, line);

			doexit(RTN_RETRY);
		}
	}

	dbg((BLOG_XFER, "got a port %d. sending recv msg...", pfd));

	(void) sprintf(line, "%c%s\n", QC_RECV, queue);
	n = strlen(line);

	if (ewrite(pfd, line, (unsigned) n, "jobsend()") != n)
		quit(0);

	dbg((BLOG_XFER, "waiting for response..."));

	/*
	 * Here we might get:
	 *	QR_ENOENT -	Remote queue not found.
	 *	QR_MT	 -	Remote queue is stopped.
	 *	QR_EAGAIN -	Remote queue not queueing.
	 *	QR_OK	-	Everything is OK.
	 */

	resp = send_get_response();

	switch(resp) {

	case QR_OK:		/* We're cool, lets send some files! */
		dbg((BLOG_XFER, "got positive response (%o)", resp));
		break;

	case QR_EAGAIN:		/* Queue is disabled */
		(void) sprintf(line, "%s: %s@%s is not available",
			       qp->q_name, queue, jp->j_shost);

		dbg((BLOG_XFER, line));
		queue_write_status(qp, line);

		send_abort();
		doexit(RTN_RETRY);
		break;

	case QR_ENOENT:		/* There is no queue by that name */
		(void) sprintf(line, "%s: queue %s@%s does not exist",
			       qp->q_name, queue, jp->j_shost);

		dbg((BLOG_XFER, line));
		queue_write_status(qp, line);

		send_abort();
		doexit(RTN_ACCESS);
		break;

	case QR_MT:		/* Queue is stopped */
		(void) sprintf(line, "%s: %s@%s is stopped",
			       qp->q_name, queue, jp->j_shost);

		dbg((BLOG_XFER, line));
		queue_write_status(qp, line);

		send_abort();
		doexit(RTN_ACCESS);
		break;

	case QR_ENOIMP:
	case QR_ERROR:
	default:
		blog(BLOG_WARNING,
		     "job_send(): bogus response (%o)", resp);
		send_abort();
		doexit(RTN_ERROR);
		break;
	}

	
	/*
	 * Send the data files first, they are not scanned by the requeuer
	 */
	jp->j_file[0] = 'd';
	tries = 0;

again:	
	switch ((rtn = job_send_file(jp->j_file, QC_RECVDF))) {
	case  RTN_OK:
		dbg((BLOG_XFER, "df sending succeeded..."));
		break;

	case RTN_RETRY:
		if (tries > 3) {
			blog(BLOG_WARNING, "df re-sending failed...");
			send_abort();
			doexit(rtn);
		}

		dbg((BLOG_XFER, "df sending failed *re*-sending..."));
		tries++;
		goto again;
		break;

	case RTN_ACCESS:
	case RTN_ERROR:
		blog(BLOG_WARNING, "df sending failed... (%d)", rtn);
		send_abort();
		doexit(rtn);
		break;
	}

	/*
	 * Send file control file last
	 */
	jp->j_file[0] = 'c';
	tries = 0;

 yetagain:
	switch ((rtn = job_send_file(jp->j_file, QC_RECVCF))) {
	case  RTN_OK:
		dbg((BLOG_XFER, "cf sending succeeded..."));

		(void) sprintf(line, "%s: sent job %d to %s@%s",
			       qp->q_name, jp->j_id, queue, jp->j_shost);

		blog(BLOG_STATS, line);
		queue_write_status(qp, line);

		send_done();
		doexit(rtn);
		break;

	case RTN_RETRY:
		if (tries > 3) {
			blog(BLOG_WARNING, "cf re-sending failed...");
			send_abort();
			doexit(rtn);
		}
		dbg((BLOG_XFER, "cf sending failed *re*-sending..."));
		tries++;
		goto yetagain;
		break;

	case RTN_ACCESS:
	case RTN_ERROR:
		blog(BLOG_WARNING, "cf sending failed... (%d)", rtn);
		send_abort();
		doexit(rtn);
		break;
	}

	/*
	 * Should not get to here
	 */
	blog(BLOG_ERR, "returning from job_send in a strange place...");

	doexit(RTN_ERROR);

	/*NOTREACHED*/
	return (-1);
}	


/*
 * Send an individual job file to the remote machine.
 * Return positive if we should try resending.
 */
static int
job_send_file(file, type)
char *	file;
int	type;
{


	char	sendbuf[BUFSIZ];
	int	f, i;
	int	resp;
	unsigned int	 amt;


	if (lstat(file, &stbuf) < 0) {
		blog(BLOG_ERR, estat, file);
		return (RTN_ACCESS);
	}

	if ((f = open(file, O_RDONLY)) < 0) {
		blog(BLOG_ERR, eopen, file);
		return (RTN_ERROR);
	}

	(void) sprintf(line, "%c%d %s\n",
		       (unsigned char) type, (int) stbuf.st_size, file);

	amt = strlen(line);

	if (ewrite(pfd, line, amt, "send_file()") != amt)
		quit(0);

	dbg((BLOG_XFER, "Sent message asking for transfer"));

	if ((resp = send_check_response("initial")) != RTN_OK) {
		(void) close(f);
		return (resp);
	}

	dbg((BLOG_XFER, "Now sending the file %s", file));

	for (i = 0; i < stbuf.st_size; i += BUFSIZ) {
		amt = BUFSIZ;

		if (i + amt > stbuf.st_size)
			amt = stbuf.st_size - i;

		if (read(f, sendbuf, amt) != amt) {
			blog(BLOG_ERR, "incomplete read of %s (%u) --- abort",
			     file, amt);
			(void) close(f);
			return (RTN_ERROR);
		}

		if (ewrite(pfd, sendbuf, amt, file) != amt) {
			(void) close(f);
			return (RTN_ERROR);
		}
	}

	(void) close(f);

	dbg((BLOG_XFER, "transmission of file %s complete", file));

	if ((resp = send_check_response("final")) != RTN_OK)
		return (resp);

	return (RTN_OK);

}


/*
 * Check the response from the receiving batchd.
 */

static int
send_check_response(msg)
char *	msg;
{

	char	resp;


	/*
	 * Here, we might get:
	 *	QR_EAGAIN	-	resource temporarily unavailable.
	 *	QR_EPERM	-	couldn't open the file.
	 *	QR_ERROR	-	?
	 *	QR_ENOIMP	-	didn't understand the message.
	 *	QR_OK		-	file opened, ready to receive.
	 */
	resp = send_get_response();

	switch(resp) {

	case QR_OK:		/* Everything is ok - continue sending file; */
		dbg((BLOG_XFER, "good %s send response (%o)", msg, (int) resp));
		return (RTN_OK);
		break;

	case QR_EAGAIN:		/* Possible temporary failure. */
		dbg((BLOG_XFER, "temporary bad %s send response (%o)", msg, (int) resp));

		return (RTN_RETRY);
		break;

	case QR_EPERM:		/* Some kind of access or fatal error. */
	case QR_ERROR:
		dbg((BLOG_XFER,
		     "bad %s send response (%o)",  msg, (int) resp));

		return (RTN_ERROR);
		break;

		/* Not implemented?  What happended here? */
	case QR_ENOIMP:
	default:
		blog(BLOG_WARNING,
		     "job_check_response(): bogus %s response (%o)",
		     msg, (int) resp);
		return (RTN_ERROR);
		break;
	}

	return (RTN_ERROR);
}

/*
 * Get a response from the receiving process.
 */
static int
send_get_response()
{

	char	resp;


	if (read(pfd, &resp, 1) != 1) {
		blog(BLOG_ERR, elost, "get_response(): read");
		quit(0);
	}

	return ( (int) resp);
}


static void
send_abort()
{

	char	cmd[2];

	cmd[0] = QC_ABORT;
	cmd[1] = '\n';

	if (ewrite(pfd, cmd, 2, "send_abort()") != 2)
		return ;

	return;
	
}


static void
send_done()
{

	char	cmd[2];


	cmd[0] = QC_DONE;
	cmd[1] = '\n';

	if (ewrite(pfd, cmd, 2, "send_done()") != 2)
		return ;

	return;
	
}

static SIGRTN
quit(sig)
int	sig;
{
	send_abort();
	dbg((BLOG_XFER, "transfer aborted by signal %d", sig));

	/* Exit with a positive error code so that the job is not removed. */
	doexit(RTN_RETRY);
}

static SIGRTN
die(sig)
int	sig;
{
	send_abort();
	dbg((BLOG_XFER, "transfer and job aborted by signal %d", sig));

	/* Exit with a negative error code so that the job is removed. */
	doexit(RTN_SIGNAL);
}

static void
doexit(code)
int	code;
{

	if (pfd != -1)
		(void) close(pfd);

	exit(code);

}
